# Copyright 2022 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli.exceptions import WrongArgFormFactorError


class ParamFormFactorVisitor(object):
    """
    This visitor's visit method will walk the input params object of the input
    shape, visiting all fields and recursing through complex fields. Any field
    encountered will be checked for form factors.
    """

    def __init__(self, form_factor):
        self._form_factor = form_factor

    def visit(self, params, shape):
        if self._form_factor is None:
            # No need to filter out any parameter if form factor is None.
            # For example: 'refdoc' command.
            return
        else:
            return self._visit(params, shape, name='')

    def _visit(self, params, shape, name):
        # self._form_factor won't be None because it is checked in visit()
        if shape.form_factors is not None and self._form_factor not in shape.form_factors:
            raise WrongArgFormFactorError(
                arg_name=name,
                form_factor=self._form_factor,
                arg_form_factors=', '.join(shape.form_factors))
        return getattr(self, '_visit_%s' % shape.type_name)(params, shape, name)

    def _visit_object(self, params, shape, name):
        visited = dict()
        for param in params:
            if param not in shape.members:
                visited[param] = params[param]
                continue
            visited[param] = self._visit(params[param],
                                         shape.members[param],
                                         '%s.%s' % (name, param))
        return visited

    def _visit_map(self, params, shape, name):
        visited = dict()
        for param in params:
            # does not support visiting key, as it is supposed to be string
            visited[param] = self._visit(params[param],
                                         shape.value,
                                         '%s.%s' % (name, param))
        return visited

    def _visit_array(self, param, shape, name):
        visited = []
        for i, item in enumerate(param):
            visited.append(self._visit(item, shape.member, '%s[%s]' % (name, i)))
        return visited

    def _visit_boolean(self, param, shape, name):
        return param

    def _visit_integer(self, param, shape, name):
        return param

    def _visit_float(self, param, shape, name):
        return param

    def _visit_double(self, param, shape, name):
        return param

    def _visit_string(self, param, shape, name):
        return param

    def _visit_datetime(self, param, shape, name):
        return param

    def _visit_blob(self, param, shape, name):
        return param



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os


import cdpcli
from cdpcli.argparser import ArgTableArgParser
from cdpcli.argprocess import unpack_argument, unpack_cli_arg
from cdpcli.arguments import create_argument_model_from_schema, CustomArgument
from cdpcli.clicommand import CLICommand
from cdpcli.compat import OrderedDict
from cdpcli.docs import generate_doc
from cdpcli.docs import OperationDocumentGenerator
from cdpcli.extensions.arguments import OverrideRequiredArgsArgument
from cdpcli.extensions.cliinputjson import add_cli_input_json
from cdpcli.extensions.generatecliskeleton import add_generate_skeleton
from cdpcli.help import HelpCommand
from cdpcli.validate import validate_parameters


_open = open


class _FromFile(object):

    def __init__(self, *paths, **kwargs):
        """
        ``**kwargs`` can contain a ``root_module`` argument
        that contains the root module where the file contents
        should be searched.  This is an optional argument, and if
        no value is provided, will default to ``cdpcli``.  This means
        that by default we look for examples in the ``cdpcli`` module.
        """
        self.filename = None
        if paths:
            self.filename = os.path.join(*paths)
        if 'root_module' in kwargs:
            self.root_module = kwargs['root_module']
        else:
            self.root_module = cdpcli


class BasicCommand(CLICommand):

    """Basic top level command with no subcommands.

    If you want to create a new command, subclass this and
    provide the values documented below.

    """

    # This is the name of your command, so if you want to
    # create an 'cdp mycommand ...' command, the NAME would be
    # 'mycommand'
    NAME = 'commandname'
    # This is the description that will be used for the 'help'
    # command.
    DESCRIPTION = 'describe the command'
    # This is optional, if you are fine with the default synopsis
    # (the way all the built in operations are documented) then you
    # can leave this empty.
    SYNOPSIS = ''
    # If you want to provide some hand written examples, you can do
    # so here.  This is written in RST format.  This is optional,
    # you don't have to provide any examples, though highly encouraged!
    EXAMPLES = ''
    # If your command has arguments, you can specify them here.  This is
    # somewhat of an implementation detail, but this is a list of dicts
    # where the dicts match the kwargs of the CustomArgument's __init__.
    # For example, if I want to add a '--argument-one' and an
    # '--argument-two' command, I'd say:
    #
    # ARG_TABLE = [
    #     {'name': 'argument-one', 'help_text': 'This argument does foo bar.',
    #      'action': 'store', 'required': False, 'cli_type_name': 'string',},
    #     {'name': 'argument-two', 'help_text': 'This argument does some other thing.',
    #      'action': 'store', 'choices': ['a', 'b', 'c']},
    # ]
    #
    ARG_TABLE = []
    # If you want the command to have subcommands, you can provide a list of
    # dicts.  We use a list here because we want to allow a user to provide
    # the order they want to use for subcommands.
    # SUBCOMMANDS = [
    #     {'name': 'subcommand1', 'command_class': SubcommandClass},
    #     {'name': 'subcommand2', 'command_class': SubcommandClass2},
    # ]
    # The command_class must subclass from ``BasicCommand``.
    SUBCOMMANDS = []

    FROM_FILE = _FromFile
    # You can set the DESCRIPTION, SYNOPSIS, and EXAMPLES to FROM_FILE
    # and we'll automatically read in that data from the file.
    # This is useful if you have a lot of content and would prefer to keep
    # the docs out of the class definition.  For example:
    #
    # DESCRIPTION = FROM_FILE
    #
    # will set the DESCRIPTION value to the contents of
    # cdpcli/examples/<command name>/_description.rst
    # The naming conventions for these attributes are:
    #
    # DESCRIPTION = cdpcli/examples/<command name>/_description.rst
    # SYNOPSIS = cdpcli/examples/<command name>/_synopsis.rst
    # EXAMPLES = cdpcli/examples/<command name>/_examples.rst
    #
    # You can also provide a relative path and we'll load the file
    # from the specified location:
    #
    # DESCRIPTION = cdpcli/examples/<filename>
    #
    # For example:
    #
    # DESCRIPTION = FROM_FILE('command, 'subcommand, '_description.rst')
    # DESCRIPTION = 'cdpcli/examples/command/subcommand/_description.rst'
    #
    # At this point, the only other thing you have to implement is a _run_main
    # method (see the method for more information).

    def __init__(self):
        self._arg_table = None
        self._subcommand_table = None
        self._lineage = [self]

    def __call__(self, client_creator, args, parsed_globals):
        # args is the remaining unparsed args.
        # We might be able to parse these args so we need to create
        # an arg parser and parse them.
        self._subcommand_table = self._build_subcommand_table()
        self._arg_table = self._build_arg_table()
        self._handle_override_required_args(args)
        parser = ArgTableArgParser(self.arg_table, command_table=self.subcommand_table)
        parsed_args, remaining = parser.parse_known_args(args)

        # Unpack arguments
        for key, value in vars(parsed_args).items():
            cli_argument = None

            # Convert the name to use dashes instead of underscore
            # as these are how the parameters are stored in the
            # `arg_table`.
            xformed = key.replace('_', '-')
            if xformed in self.arg_table:
                cli_argument = self.arg_table[xformed]

            value = unpack_argument(cli_argument, value)

            # If this parameter has a schema defined, then allow plugins
            # a chance to process and override its value.
            if self._should_allow_override(cli_argument, value):
                # Unpack the argument, which is a string, into the
                # correct Python type (dict, list, etc)
                value = unpack_cli_arg(cli_argument, value)
                self._validate_value_against_schema(
                    cli_argument.argument_model, value)

            setattr(parsed_args, key, value)

        if hasattr(parsed_args, 'help'):
            self._display_help(client_creator, parsed_args, parsed_globals)
        elif getattr(parsed_args, 'subcommand', None) is None:
            # No subcommand was specified so call the main
            # function for this top level command.
            if remaining:
                raise ValueError("Unknown options: %s" % ','.join(remaining))
            return self._run_main(client_creator,
                                  parsed_args,
                                  parsed_globals)
        else:
            return self.subcommand_table[parsed_args.subcommand](client_creator,
                                                                 remaining,
                                                                 parsed_globals)

    def _validate_value_against_schema(self, model, value):
        validate_parameters(value, model)

    def _handle_override_required_args(self, args):
        argument_table = self.arg_table
        for cli_name, cli_argument in argument_table.items():
            if isinstance(cli_argument, OverrideRequiredArgsArgument):
                cli_argument.override_required_args(argument_table, args)
                self._operation_callers.insert(0, cli_argument)

    def _should_allow_override(self, param, value):
        if (param and param.argument_model is not None and
                value is not None):
            return True
        return False

    def _run_main(self, client_creator, parsed_args, parsed_globals):
        # Subclasses should implement this method.
        # parsed_globals are the parsed global args (things like region,
        # profile, output, etc.)
        # context holds the configuration for the application.
        # parsed_args are any arguments you've defined in your ARG_TABLE
        # that are parsed.  These will come through as whatever you've
        # provided as the 'dest' key.  Otherwise they default to the
        # 'name' key.  For example: ARG_TABLE[0] = {"name": "foo-arg", ...}
        # can be accessed by ``parsed_args.foo_arg``.
        raise NotImplementedError("_run_main")

    def _build_subcommand_table(self):
        subcommand_table = OrderedDict()
        for subcommand in self.SUBCOMMANDS:
            subcommand_name = subcommand['name']
            subcommand_class = subcommand['command_class']
            subcommand_table[subcommand_name] = subcommand_class()
        self._add_lineage(subcommand_table)
        return subcommand_table

    def _display_help(self, client_creator, parsed_args, parsed_globals):
        help_command = self.create_help_command()
        help_command(client_creator, parsed_args, parsed_globals)

    def create_help_command(self):
        command_help_table = {}
        if self.SUBCOMMANDS:
            command_help_table = self.create_help_command_table()
        return BasicHelp(self,
                         command_table=command_help_table,
                         arg_table=self.arg_table,
                         command_lineage='.'.join(self.lineage_names))

    def create_help_command_table(self):
        """
        Create the command table into a form that can be handled by the
        BasicDocHandler.
        """
        commands = {}
        for command in self.SUBCOMMANDS:
            commands[command['name']] = command['command_class']()
        self._add_lineage(commands)
        return commands

    def _build_arg_table(self):
        arg_table = OrderedDict()
        argument_model = None
        for arg_data in self.ARG_TABLE:
            # If a custom schema was passed in, create the argument_model
            # so that it can be validated and docs can be generated.
            if 'schema' in arg_data:
                argument_model = create_argument_model_from_schema(
                    arg_data.pop('schema'))
                arg_data['argument_model'] = argument_model
            custom_argument = CustomArgument(**arg_data)

            arg_table[arg_data['name']] = custom_argument
        if argument_model:
            add_cli_input_json(argument_model, arg_table)
            add_generate_skeleton(argument_model, arg_table)
            self.argument_model = argument_model
        return arg_table

    def _add_lineage(self, command_table):
        for command in command_table:
            command_obj = command_table[command]
            command_obj.lineage = self.lineage + [command_obj]

    @property
    def arg_table(self):
        if self._arg_table is None:
            self._arg_table = self._build_arg_table()
        return self._arg_table

    @property
    def subcommand_table(self):
        if self._subcommand_table is None:
            self._subcommand_table = self._build_subcommand_table()
        return self._subcommand_table

    @classmethod
    def add_command(cls, command_table):
        command_table[cls.NAME] = cls()

    @property
    def name(self):
        return self.NAME

    @property
    def lineage(self):
        return self._lineage

    @lineage.setter
    def lineage(self, value):
        self._lineage = value


class BasicDocHandler(OperationDocumentGenerator):

    def __init__(self, help_command):
        super(BasicDocHandler, self).__init__(help_command)
        self.doc = help_command.doc

    def build_translation_map(self):
        return {}

    def doc_description(self, help_command, **kwargs):
        self.doc.style.h2('Description')
        self.doc.write(help_command.description)
        self.doc.style.new_paragraph()

    def doc_synopsis_start(self, help_command):
        if not help_command.synopsis:
            super(BasicDocHandler, self).doc_synopsis_start(
                help_command=help_command)
        else:
            self.doc.style.h2('Synopsis')
            self.doc.style.start_codeblock()
            self.doc.writeln(help_command.synopsis)

    def doc_synopsis_option(self, arg_name, help_command, **kwargs):
        if not help_command.synopsis:
            doc = help_command.doc
            argument = help_command.arg_table[arg_name]
            if argument.synopsis:
                option_str = argument.synopsis
            elif argument.group_name in self._arg_groups:
                if argument.group_name in self._documented_arg_groups:
                    # This arg is already documented so we can move on.
                    return
                option_str = ' | '.join(
                    [a.cli_name for a in
                     self._arg_groups[argument.group_name]])
                self._documented_arg_groups.append(argument.group_name)
            elif argument.cli_type_name == 'boolean':
                option_str = '%s' % argument.cli_name
            elif argument.nargs == '+':
                option_str = "%s <value> [<value>...]" % argument.cli_name
            else:
                option_str = '%s <value>' % argument.cli_name
            if not (argument.required or argument.positional_arg):
                option_str = '[%s]' % option_str
            doc.writeln('%s' % option_str)

        else:
            # A synopsis has been provided so we don't need to write
            # anything here.
            pass

    def doc_synopsis_end(self, help_command):
        if not help_command.synopsis:
            super(BasicDocHandler, self).doc_synopsis_end(
                help_command=help_command)
        else:
            self.doc.style.end_codeblock()

    def doc_examples(self, help_command):
        if help_command.examples:
            self.doc.style.h2('Examples')
            self.doc.write(help_command.examples)

    def doc_subitems_start(self, help_command):
        if help_command.command_table:
            doc = help_command.doc
            doc.style.h2('Available Commands')
            doc.style.toctree()

    def doc_subitem(self, command_name, help_command):
        if help_command.command_table:
            doc = help_command.doc
            doc.style.tocitem(command_name)

    def doc_subitems_end(self, help_command):
        pass

    def doc_output(self, help_command):
        pass

    def doc_option_example(self, arg_name, help_command):
        pass


class BasicHelp(HelpCommand):
    GeneratorClass = BasicDocHandler

    def __init__(self,
                 command_object,
                 command_table,
                 arg_table,
                 command_lineage):
        super(BasicHelp, self).__init__(command_object,
                                        command_table,
                                        arg_table)

        # These are public attributes that are mapped from the command
        # object.
        self._description = command_object.DESCRIPTION
        self._synopsis = command_object.SYNOPSIS
        self._examples = command_object.EXAMPLES
        self._command_lineage = command_lineage

    @property
    def command_lineage(self):
        return self._command_lineage

    @property
    def name(self):
        return self.obj.NAME

    @property
    def description(self):
        return self._get_doc_contents('_description')

    @property
    def synopsis(self):
        return self._get_doc_contents('_synopsis')

    @property
    def examples(self):
        return self._get_doc_contents('_examples')

    def _get_doc_contents(self, attr_name):
        value = getattr(self, attr_name)
        if isinstance(value, BasicCommand.FROM_FILE):
            if value.filename is not None:
                trailing_path = value.filename
            else:
                trailing_path = os.path.join(self.name, attr_name + '.rst')
            root_module = value.root_module
            doc_path = os.path.join(
                os.path.abspath(os.path.dirname(root_module.__file__)),
                'examples', trailing_path)
            with _open(doc_path) as f:
                return f.read()
        else:
            return value

    def __call__(self, client_creator, args, parsed_globals):
        # Now generate all of the events for a Provider document.
        # We pass ourselves along so that we can, in turn, get passed
        # to all event handlers.
        generate_doc(self.GeneratorClass(self), self)
        self.renderer.render(self.doc.getvalue())



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli.compat import json
import dateutil.parser


class ResponseParserFactory(object):

    def create_parser(self):
        return ResponseParser()


class ResponseParser(object):
    DEFAULT_ENCODING = 'utf-8'

    def parse(self, response, shape):
        if response['status_code'] >= 301:
            return self._parse_error(response)
        elif 'Content-Type' in response['headers'] and \
             not self._is_json_response(response['headers']['Content-Type']):
            return response['body']
        elif shape is None:
            return {}
        else:
            return self._parse_shape(shape, self._decode_body(response))

    def _is_json_response(self, content_type):
        if not content_type:
            return False
        if content_type == 'application/json':
            return True
        if content_type.startswith('application/json;'):
            return True
        return False

    def _decode_body(self, response):
        body = response['body']
        if not body:
            return {}
        return json.loads(body.decode(self.DEFAULT_ENCODING))

    def _parse_error(self, response):
        try:
            body = self._decode_body(response)
        except Exception:
            # In the case that we hit a completely dead endpoint, etc, the
            # response might not be valid JSON, but we still want to return
            # a structured error to the caller.
            message = response['body'].decode(self.DEFAULT_ENCODING).strip()
            body = dict(code='UNKNOWN_ERROR', message=message)
        error = {}
        error['code'] = body.get('code', '')
        if 'message' in body:
            error['message'] = body.get('message', '')
        elif 'error' in body:
            error['message'] = body.get('error', '')
        else:
            error['message'] = ' '.join(body.get('errorMessages', []))
        return {'error': error}

    def _parse_shape(self, shape, value):
        handle_method_name = '_handle_%s' % shape.type_name
        method = getattr(self, handle_method_name, self._handle_default)
        return method(shape, value)

    def _handle_array(self, shape, value):
        parsed = []
        for item in value:
            parsed.append(self._parse_shape(shape.member, item))
        return parsed

    def _handle_object(self, shape, value):
        parsed = {}
        for member_name in shape.members:
            member_shape = shape.members[member_name]
            member_value = value.get(member_shape.name)
            if member_value is not None:
                parsed[member_name] = self._parse_shape(
                    shape.members[member_name], member_value)
        return parsed

    def _handle_map(self, shape, values):
        parsed = {}
        for key, value in values.items():
            actual_key = self._parse_shape(shape.key, key)
            actual_value = self._parse_shape(shape.value, value)
            parsed[actual_key] = actual_value
        return parsed

    # Handles both ISO8601 format and RFC822 format.
    # However, server can only parse ISO8601 format as of now.
    def _handle_datetime(self, shape, value):
        try:
            return dateutil.parser.parse(value)
        except (TypeError, ValueError) as e:
            raise ValueError('Invalid timestamp "%s": %s' % (value, e))

    def _handle_default(self, shape, value):
        return value



# Copyright 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import http.server as httpserver
import os
import socket
import socketserver
import sys
import urllib.parse as urlparse
import webbrowser

from cdpcli import CDP_ACCESS_KEY_ID_KEY_NAME, CDP_PRIVATE_KEY_KEY_NAME
from cdpcli.endpoint import EndpointResolver
from cdpcli.exceptions import InteractiveLoginError, MissingArgumentError, ProfileNotFound
from cdpcli.extensions.commands import BasicCommand
from cdpcli.extensions.configure import CREDENTIAL_FILE_COMMENT
from cdpcli.extensions.writer import ConfigFileWriter


DEFAULT_LOGIN_TIMEOUT = 600


class LoginHttpServer(socketserver.TCPServer):
    """
    A HTTP-server to listen and accept then access-key when interactive login succeeded.
    """

    def __init__(self, port, context, config_writer):
        server_address = ('localhost', port)
        super(LoginHttpServer, self).__init__(server_address, LoginHttpHandler)
        self.context = context
        self.config_writer = config_writer

    def handle_timeout(self):
        """
        Raise an exception if timeout.
        The caller of this HTTP-server will catch the exception and stop the server.
        """
        raise InteractiveLoginError(err_msg='Login timeout')


class LoginHttpHandler(httpserver.SimpleHTTPRequestHandler):
    """
    A HTTP-handler to handle following request:
    http://localhost/interactiveLogin?accessKeyId=...&privateKey=...
    1) save the token.
    2) send CLOSE_BROWER_HTML as response.
    """

    CLOSE_BROWSER_HTML = b'''\
<!DOCTYPE html>
<html>
  <head>
    <link rel="icon" href="data:,">
    <meta charset="utf-8" />
  </head>
  <body onload="window.close()">
    <p>It is safe to close your browser.</p>
  </body>
</html>
'''

    def log_message(self, format, *args):
        pass

    def do_GET(self):
        if not self.path.startswith('/interactiveLogin'):
            sys.stderr.write('Login failed: URL path not supported\n')
            self._send_response(404, b'')
            return

        url_parsed = urlparse.urlsplit(self.path)
        url_params = urlparse.parse_qs(url_parsed.query)

        error = url_params.get('error')
        if error is not None:
            if isinstance(error, list):
                error = error[0]
            # print the error message
            sys.stderr.write('Login failed: %s\n' % error)
            # send empty response, do not close the browser
            self._send_response(200, b'')
            return

        access_key_id = url_params.get('accessKeyId')
        private_key = url_params.get('privateKey')

        # access_key_id and private_key are lists (url params are lists by default),
        # read the first element in the list.
        if isinstance(access_key_id, list):
            access_key_id = access_key_id[0]
        if isinstance(private_key, list):
            private_key = private_key[0]
        if access_key_id is None or private_key is None:
            sys.stderr.write('Login failed: Missing access key id or private key\n')
            self._send_response(400, b'Missing access key id or private key')
            return

        # save the access token.
        self._save_access_token(access_key_id, private_key)
        # send close browser HTML to the client.
        self._send_response(200, LoginHttpHandler.CLOSE_BROWSER_HTML)

    def _send_response(self, status_code, body):
        self.send_response(status_code)
        self.end_headers()
        self.wfile.write(body)

    def _save_access_token(self, access_key_id, private_key):
        credential_file_values = {CDP_ACCESS_KEY_ID_KEY_NAME: access_key_id,
                                  CDP_PRIVATE_KEY_KEY_NAME: private_key}
        profile_name = self.server.context.effective_profile
        if profile_name is not None:
            credential_file_values['__section__'] = profile_name
        shared_credentials_filename = os.path.expanduser(
            self.server.context.get_config_variable('credentials_file'))
        self.server.config_writer.update_config(
            credential_file_values,
            shared_credentials_filename,
            config_file_comment=CREDENTIAL_FILE_COMMENT)


class LoginCommand(BasicCommand):
    """
    The 'cdp login' command handler.
    """

    NAME = 'login'
    DESCRIPTION = 'Login to CDP interactively.'
    SYNOPSIS = (
        'cdp login'
        ' [--account-id account-id]'
        ' [--identity-provider identity-provider-name]'
    )
    EXAMPLES = (
        'To login interactively::\n'
        '\n'
        '    $ cdp login --account-id <guid>\n'
    )
    SUBCOMMANDS = []
    ARG_TABLE = [
        {'name': 'account-id',
         'help_text': ('The account-id of the tenant to login. This is a required '
                       'parameter. \'cdp configure set account_id <account-id>\' '
                       'could be used to set the default account-id to be used if '
                       'this parameter is not provided.'),
         'action': 'store',
         'required': False,  # the account id could also come from config file.
         'cli_type_name': 'string'},
        {'name': 'identity-provider',
         'help_text': ('The name or CRN of IdP which will be used to authenticate users. '
                       'The default IdP will be used if not provided.'),
         'action': 'store',
         'required': False,
         'cli_type_name': 'string'},
        {'name': 'login-url',
         'help_text': ('The URL (SP-initiated or IdP-initiated URL) to '
                       'start the interactive login with.'),
         'action': 'store',
         'required': False,
         'hidden': True,
         'no_paramfile': True,
         'cli_type_name': 'string'},
        {'name': 'port',
         'help_text': ('The listening port number for CLI to receive the access token. '
                       'A random un-used port will be assigned if not provided.'),
         'action': 'store',
         'required': False,
         'hidden': True,
         'cli_type_name': 'integer'},
        {'name': 'timeout',
         'help_text': ('The login timeout duration in seconds. Default to '
                       '%d seconds if not provided.') % DEFAULT_LOGIN_TIMEOUT,
         'action': 'store',
         'required': False,
         'hidden': True,
         'cli_type_name': 'integer'},
    ]

    def __init__(self, config_writer=None, open_new_browser=None):
        super(LoginCommand, self).__init__()
        if config_writer is None:
            config_writer = ConfigFileWriter()
        self._config_writer = config_writer
        if open_new_browser is None:
            open_new_browser = LoginCommand._open_new_browser
        self._open_new_browser = open_new_browser

    def _run_main(self, client_creator, parsed_args, parsed_globals):
        # Called when invoked with no sub-command "cdp login"

        # This is the config from the config file scoped to a specific
        # profile.
        try:
            context = client_creator.context
            config = context.get_scoped_config()
        except ProfileNotFound:
            config = {}

        port = parsed_args.port
        if port is None:
            port = LoginCommand._find_unused_port()
        timeout = parsed_args.timeout
        if timeout is None:
            timeout = DEFAULT_LOGIN_TIMEOUT

        login_url = self._resolve_login_url(parsed_args, parsed_globals, config, port)
        self._open_new_browser(login_url)
        self._run_http_server(port, timeout, context)

    @staticmethod
    def _open_new_browser(url):
        webbrowser.open_new(url)

    @staticmethod
    def _find_unused_port():
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(('localhost', 0))
        port = s.getsockname()[1]
        s.close()
        return port

    def _resolve_login_url(self, parsed_args, parsed_globals, config, return_url_port):
        # get the login URL from input parameter or config file.
        endpoint_resolver = EndpointResolver()
        login_url = endpoint_resolver.resolve(
            explicit_endpoint_url=parsed_args.login_url,
            config=config,
            region=parsed_globals.cdp_region,
            service_name='LOGIN',
            prefix=None,
            products=['CDP'],
            scheme='https',
            port=443)

        # get the account-id and idp from:
        # 1) login URL
        # 2) input parameter
        # 3) config file
        url_parsed = urlparse.urlsplit(login_url)
        url_params = urlparse.parse_qs(url_parsed.query)
        if 'accountId' not in url_params:
            account_id = parsed_args.account_id
            if account_id is None:
                account_id = config.get('account_id')
            if account_id is not None:
                url_params['accountId'] = account_id
            else:
                raise MissingArgumentError(arg_name='--account-id')
        if 'idp' not in url_params:
            idp = parsed_args.identity_provider
            if idp is None:
                idp = config.get('identity_provider')
            if idp is not None:
                url_params['idp'] = idp

        return_url = 'http://localhost:%d/interactiveLogin' % return_url_port
        url_params['returnUrl'] = return_url
        url_query = urlparse.urlencode(url_params, doseq=True)
        url_parsed = url_parsed._replace(query=url_query)
        return urlparse.urlunsplit(url_parsed)

    def _run_http_server(self, port, timeout, context):
        httpd = LoginHttpServer(port, context, self._config_writer)
        try:
            # handle_request() waits and handles one HTTP request.
            httpd.timeout = timeout
            httpd.handle_request()
        finally:
            httpd.server_close()



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli import LIST_TYPE
from cdpcli import MAP_TYPE
from cdpcli import OBJECT_TYPE
from cdpcli.argprocess import ParamShorthand
from cdpcli.argprocess import unpack_cli_arg
from cdpcli.model import ShapeResolver
from cdpcli.schema import SchemaTransformer


def create_argument_model_from_schema(schema):
    # Given a JSON schema (described in schema.py), convert it
    # to a shape object from `cdp.model.Shape` that can be
    # used as the argument_model for the Argument classes below.
    transformer = SchemaTransformer()
    shapes_map = transformer.transform(schema)
    shape_resolver = ShapeResolver(shapes_map)
    # The SchemaTransformer guarantees that the top level shape
    # will always be named 'InputShape'.
    arg_shape = shape_resolver.get_shape_by_name('InputShape', 'InputShape')
    return arg_shape


class UnknownArgumentError(Exception):
    pass


def first_non_none_response(responses, default=None):
    for response in responses:
        if response[1] is not None:
            return response[1]
    return default


class BaseCLIArgument(object):

    def __init__(self, name):
        self._name = name

    def add_to_arg_table(self, argument_table):
        argument_table[self.name] = self

    def add_to_parser(self, parser):
        pass

    def add_to_params(self, parameters, value):
        pass

    @property
    def name(self):
        return self._name

    @property
    def cli_name(self):
        return '--' + self._name

    @property
    def cli_type_name(self):
        raise NotImplementedError("cli_type_name")

    @property
    def required(self):
        raise NotImplementedError("required")

    @property
    def documentation(self):
        raise NotImplementedError("documentation")

    @property
    def cli_type(self):
        raise NotImplementedError("cli_type")

    @property
    def py_name(self):
        return self._name.replace('-', '_')

    @property
    def choices(self):
        return None

    @property
    def synopsis(self):
        return ''

    @property
    def positional_arg(self):
        return False

    @property
    def nargs(self):
        return None

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def group_name(self):
        return None


class CustomArgument(BaseCLIArgument):

    def __init__(self,
                 name,
                 help_text='',
                 dest=None,
                 default=None,
                 action=None,
                 required=None,
                 choices=None,
                 nargs=None,
                 cli_type_name=None,
                 group_name=None,
                 positional_arg=False,
                 no_paramfile=False,
                 argument_model=None,
                 synopsis='',
                 hidden=False):
        self._name = name
        self._help = help_text
        self._dest = dest
        self._default = default
        self._action = action
        self._required = required
        self._nargs = nargs
        self._cli_type_name = cli_type_name
        self._group_name = group_name
        self._positional_arg = positional_arg
        if choices is None:
            choices = []
        self._choices = choices
        self._synopsis = synopsis
        self._UNDOCUMENTED = hidden

        # These are public attributes that are ok to access from external
        # objects.
        self.no_paramfile = no_paramfile
        if argument_model is None:
            argument_model = self._create_scalar_argument_model()
        self.argument_model = argument_model

        # If the top level element is a list then set nargs to
        # accept multiple values seperated by a space.
        if self.argument_model is not None and \
                self.argument_model.type_name == LIST_TYPE:
            self._nargs = '+'

    def _create_scalar_argument_model(self):
        if self._nargs is not None:
            # If nargs is not None then argparse will parse the value
            # as an array, so we don't create an argument_object so we don't
            # go through param validation.
            return None
        # If no argument model is provided, we create a basic
        # shape argument.
        type_name = self.cli_type_name
        return create_argument_model_from_schema({'type': type_name})

    @property
    def cli_name(self):
        if self._positional_arg:
            return self._name
        else:
            return '--' + self._name

    def add_to_parser(self, parser):
        cli_name = self.cli_name
        kwargs = {}
        if self._dest is not None:
            kwargs['dest'] = self._dest
        if self._action is not None:
            kwargs['action'] = self._action
        if self._default is not None:
            kwargs['default'] = self._default
        if self._choices:
            kwargs['choices'] = self._choices
        if self._required is not None:
            kwargs['required'] = self._required
        if self._nargs is not None:
            kwargs['nargs'] = self._nargs
        parser.add_argument(cli_name, **kwargs)

    @property
    def required(self):
        if self._required is None:
            return False
        return self._required

    @required.setter
    def required(self, value):
        self._required = value

    @property
    def documentation(self):
        return self._help

    @property
    def cli_type_name(self):
        if self._cli_type_name is not None:
            return self._cli_type_name
        elif self._action in ['store_true', 'store_false']:
            return 'boolean'
        else:
            # Default to 'string' type if we don't have any
            # other info.
            return 'string'

    @property
    def cli_type(self):
        cli_type = str
        if self._action in ['store_true', 'store_false']:
            cli_type = bool
        return cli_type

    @property
    def choices(self):
        return self._choices

    @property
    def group_name(self):
        return self._group_name

    @property
    def synopsis(self):
        return self._synopsis

    @property
    def positional_arg(self):
        return self._positional_arg

    @property
    def nargs(self):
        return self._nargs


class CLIArgument(BaseCLIArgument):
    """Represents a CLI argument that maps to a service parameter.

    """

    TYPE_MAP = {
        OBJECT_TYPE: str,
        MAP_TYPE: str,
        'timestamp': str,
        LIST_TYPE: str,
        'string': str,
        'float': float,
        'integer': str,
        'long': int,
        'boolean': bool,
        'double': float,
        'blob': str
    }

    def __init__(self,
                 name,
                 argument_model,
                 operation_model,
                 is_required=False,
                 serialized_name=None,
                 no_paramfile=False):
        self._name = name
        # This is the name we need to use when constructing the parameters
        # dict we send to CDP.  While we can change the .name attribute
        # which is the name exposed in the CLI, the serialized name we use
        # for CDP is invariant and should not be changed.
        if serialized_name is None:
            serialized_name = name
        self._serialized_name = serialized_name
        self.argument_model = argument_model
        self._required = is_required
        self._operation_model = operation_model
        self.no_paramfile = no_paramfile
        self._UNDOCUMENTED = self.argument_model.is_undocumented

    @property
    def py_name(self):
        return self._name.replace('-', '_')

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, value):
        self._required = value

    @property
    def documentation(self):
        return self.argument_model.documentation

    @property
    def cli_type_name(self):
        return self.argument_model.type_name

    @property
    def cli_type(self):
        return self.TYPE_MAP.get(self.argument_model.type_name, str)

    def add_to_parser(self, parser):
        cli_name = self.cli_name
        parser.add_argument(
            cli_name,
            help=self.documentation,
            type=self.cli_type,
            required=self.required)

    def add_to_params(self, parameters, value):
        if value is None:
            return
        else:
            # This is a two step process.  First is the process of converting
            # the command line value into a python value.  Normally this is
            # handled by argparse directly, but there are cases where extra
            # processing is needed.  For example, "--foo name=value" the value
            # can be converted from "name=value" to {"name": "value"}.  This is
            # referred to as the "unpacking" process.  Once we've unpacked the
            # argument value, we have to decide how this is converted into
            # something that can be consumed by CDP.  Many times this is
            # just associating the key and value in the params dict as down
            # below.  Sometimes this can be more complicated, and subclasses
            # can customize as they need.
            unpacked = self._unpack_argument(value)
            parameters[self._serialized_name] = unpacked

    def _unpack_argument(self, value):
        override = ParamShorthand()(self, value)
        if override is not None:
            return override
        return unpack_cli_arg(self, value)


class ListArgument(CLIArgument):

    def add_to_parser(self, parser):
        cli_name = self.cli_name
        parser.add_argument(cli_name,
                            nargs='*',
                            type=self.cli_type,
                            required=self.required)


class BooleanArgument(CLIArgument):

    def __init__(self,
                 name,
                 argument_model,
                 operation_model,
                 is_required=False,
                 action='store_true',
                 dest=None,
                 group_name=None,
                 default=None,
                 serialized_name=None,
                 no_paramfile=False):
        super(BooleanArgument, self).__init__(name,
                                              argument_model,
                                              operation_model,
                                              is_required,
                                              serialized_name=serialized_name,
                                              no_paramfile=no_paramfile)
        self._mutex_group = None
        self._action = action
        if dest is None:
            self._destination = self.py_name
        else:
            self._destination = dest
        if group_name is None:
            self._group_name = self.name
        else:
            self._group_name = group_name
        self._default = default

    def add_to_params(self, parameters, value):
        # If a value was explicitly specified (so value is True/False
        # but *not* None) then we add it to the params dict.
        # If the value was not explicitly set (value is None)
        # we don't add it to the params dict.
        if value is not None:
            parameters[self._serialized_name] = value

    def add_to_arg_table(self, argument_table):
        # Boolean parameters are a bit tricky.  For a single boolean parameter
        # we actually want two CLI params, a --foo, and a --no-foo.  To do this
        # we need to add two entries to the argument table.  So we can add
        # ourself as the positive option (--no), and then create a clone of
        # ourselves for the negative service.  We then insert both into the
        # arg table.
        argument_table[self.name] = self
        negative_name = 'no-%s' % self.name
        negative_version = self.__class__(negative_name,
                                          self.argument_model,
                                          self._operation_model,
                                          action='store_false',
                                          dest=self._destination,
                                          group_name=self.group_name,
                                          serialized_name=self._serialized_name)
        argument_table[negative_name] = negative_version

    def add_to_parser(self, parser):
        parser.add_argument(self.cli_name,
                            help=self.documentation,
                            action=self._action,
                            default=self._default,
                            dest=self._destination)

    @property
    def group_name(self):
        return self._group_name



# Copyright (c) 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli.exceptions import RedirectExtensionError
from cdpcli.extensions.redirect import Redirect
from mock import Mock
from tests import unittest


class TestRedirectExtension(unittest.TestCase):

    def setUp(self):
        self.client = Mock()
        self.client.raise_error.side_effect = Exception('ClientError')
        self.client_creator = Mock(return_value=self.client)
        self.operation_model = Mock()
        self.operation_model.service_model = Mock(service_name='service-name')
        self.operation_model.name = 'operation'
        self.operation_model.http = {'method': 'post', 'requestUri': None}
        self.parsed_globals = Mock(output=None)

    def test_redirect(self):
        parsed_globals = Mock(output=None)
        redirect_http = Mock(status_code=308, is_redirect=True)
        redirect_http.headers = {'Location': 'https://test.com/path/uploadFlow'}
        self.client.make_api_call.return_value = (redirect_http, None)
        self.client.make_request.return_value = (Mock(status_code=200), {})

        parameters = {'name': 'foo'}
        redirect = Redirect()
        redirect.invoke(self.client_creator, self.operation_model,
                        parameters, None, parsed_globals)

        self.assertEqual(1, self.client.make_api_call.call_count)
        args, kwargs = self.client.make_api_call.call_args
        self.assertEqual('operation', args[0])
        self.assertEqual(parameters, args[1])
        self.assertFalse(kwargs['allow_redirects'])

        self.assertEquals('https://test.com/path/uploadFlow',
                          parsed_globals.endpoint_url)
        self.assertEquals('https://test.com/path/uploadFlow',
                          self.operation_model.http['requestUri'])

    def test_no_redirect(self):
        redirect_http = Mock(status_code=200, is_redirect=False)
        redirect_http.headers = {'Location': 'https://test.com/path/uploadFlow'}
        self.client.make_api_call.return_value = (redirect_http, None)

        parameters = {'name': 'foo'}
        redirect = Redirect()
        with self.assertRaisesRegex(RedirectExtensionError,
                                    'The redirect extension failed: '
                                    'Missing or incorrect redirect URL.'):
            redirect.invoke(self.client_creator, self.operation_model,
                            parameters, None, self.parsed_globals)
        self.assertEqual(1, self.client.make_api_call.call_count)
        self.assertEqual(0, self.client.make_request.call_count)
        self.assertIsNone(self.operation_model.http['requestUri'])

    def test_no_redirect_url(self):
        redirect_http = Mock(status_code=308, is_redirect=True)
        redirect_http.headers = {}
        self.client.make_api_call.return_value = (redirect_http, None)

        parameters = {'name': 'foo'}
        redirect = Redirect()
        with self.assertRaisesRegex(RedirectExtensionError,
                                    'The redirect extension failed: '
                                    'Missing or incorrect redirect URL.'):
            redirect.invoke(self.client_creator, self.operation_model,
                            parameters, None, self.parsed_globals)
        self.assertEqual(1, self.client.make_api_call.call_count)
        self.assertEqual(0, self.client.make_request.call_count)
        self.assertIsNone(self.operation_model.http['requestUri'])

    def test_incorrect_redirect_url(self):
        redirect_http = Mock(status_code=308, is_redirect=True)
        redirect_http.headers = {'Location': '/relative-path/flows'}
        self.client.make_api_call.return_value = (redirect_http, None)

        parameters = {'name': 'foo'}
        redirect = Redirect()
        with self.assertRaisesRegex(RedirectExtensionError,
                                    'The redirect extension failed: '
                                    'Missing or incorrect redirect URL.'):
            redirect.invoke(self.client_creator, self.operation_model,
                            parameters, None, self.parsed_globals)
        self.assertEqual(1, self.client.make_api_call.call_count)
        self.assertEqual(0, self.client.make_request.call_count)
        self.assertIsNone(self.operation_model.http['requestUri'])



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections import namedtuple
import logging
import os

from cdpcli import CDP_ACCESS_KEY_ID_KEY_NAME, \
                   CDP_ACCESS_TOKEN_KEY_NAME, \
                   CDP_PRIVATE_KEY_KEY_NAME
from cdpcli.auth import AccessTokenAuth
from cdpcli.auth import Ed25519v1Auth
import cdpcli.compat
from cdpcli.compat import json
from cdpcli.configloader import raw_config_parse
from cdpcli.exceptions import ConfigNotFound
from cdpcli.exceptions import MalformedCredentialsError
from cdpcli.exceptions import NoCredentialsError
from cdpcli.exceptions import PartialCredentialsError
from cdpcli.exceptions import UnknownCredentialError

LOG = logging.getLogger('cdpcli.credentials')
ReadOnlyCredentials = namedtuple('ReadOnlyCredentials',
                                 ['access_key_id',
                                  'private_key',
                                  'access_token',
                                  'method'])
ACCESS_KEY_ID = 'access_key_id'
PRIVATE_KEY = 'private_key'
ACCESS_TOKEN = 'access_token'


def create_credential_resolver(context, parsed_globals=None):
    """Create a default credential resolver.

    This creates a pre-configured credential resolver
    that includes the default lookup chain for
    credentials.

    :param parsed_globals: CLI input parameters which might contain an access-token.
    """
    profile_name = context.effective_profile
    auth_file = context.get_config_variable('auth_config')
    shared_credential_file = context.get_config_variable('credentials_file')

    env_provider = EnvProvider()
    providers = [
        ParamsProvider(parsed_globals),
        env_provider,
        AuthConfigFile(auth_file),
        SharedCredentialProvider(
            creds_filename=shared_credential_file,
            profile_name=profile_name
        ),
    ]

    explicit_profile = context.get_config_variable('profile',
                                                   methods=('instance',))
    if explicit_profile is not None:
        # An explicitly provided profile will negate an EnvProvider.
        # We will defer to providers that understand the "profile"
        # concept to retrieve credentials.
        # The one edge case is if all three values are provided via
        # env vars:
        # export CDP_ACCESS_KEY_ID=foo
        # export CDP_PRIVATE_KEY=bar
        # export CDP_PROFILE=baz
        # Then, just like our client() calls, the explicit credentials
        # will take precedence.
        #
        # This precedence is enforced by leaving the EnvProvider in the chain.
        # This means that the only way a "profile" would win is if the
        # EnvProvider does not return credentials, which is what we want
        # in this scenario.
        providers.remove(env_provider)
        LOG.debug('Skipping environment variable credential check because '
                  'profile name was explicitly set.')

    resolver = CredentialResolver(providers=providers)
    return resolver


def get_credentials(context, parsed_globals=None):
    resolver = create_credential_resolver(context, parsed_globals)
    return resolver.load_credentials()


class Credentials(object):
    """
    Holds the credentials needed to authenticate requests.
    """

    def __init__(self,
                 access_key_id=None,
                 private_key=None,
                 access_token=None,
                 method=None):
        self.access_key_id = access_key_id
        self.private_key = private_key
        self.access_token = access_token
        self.method = method
        self._normalize()

    def _normalize(self):
        self.access_key_id = cdpcli.compat.ensure_unicode(self.access_key_id)
        self.private_key = cdpcli.compat.ensure_unicode(self.private_key)
        self.access_token = cdpcli.compat.ensure_unicode(self.access_token)

    def get_frozen_credentials(self):
        return ReadOnlyCredentials(self.access_key_id,
                                   self.private_key,
                                   self.access_token,
                                   self.method)


class CredentialProvider(object):

    # Implementations must provide a method.
    METHOD = None

    def load(self):
        return True

    def _extract_creds_from_mapping(self, mapping, *key_names):
        found = []
        for key_name in key_names:
            try:
                found.append(mapping[key_name])
            except KeyError:
                raise PartialCredentialsError(provider=self.METHOD,
                                              cred_var=key_name)
        if len(found) == 1:  # found access-token, returns a single-value string.
            return found[0]
        else:  # found access-key-id and private-key, returns a tuple.
            return found


class EnvProvider(CredentialProvider):
    METHOD = 'env'
    ACCESS_KEY_ID_ENV_VAR = 'CDP_ACCESS_KEY_ID'
    PRIVATE_KEY_ENV_VAR = 'CDP_PRIVATE_KEY'
    ACCESS_TOKEN_ENV_VAR = 'CDP_ACCESS_TOKEN'

    def __init__(self, environ=None, mapping=None):
        super(EnvProvider, self).__init__()
        if environ is None:
            environ = os.environ
        self.environ = environ
        self._mapping = self._build_mapping(mapping)

    def _build_mapping(self, mapping):
        # Mapping of variable name to env var name.
        var_mapping = {}
        if mapping is None:
            # Use the class var default.
            var_mapping[ACCESS_KEY_ID] = self.ACCESS_KEY_ID_ENV_VAR
            var_mapping[PRIVATE_KEY] = self.PRIVATE_KEY_ENV_VAR
            var_mapping[ACCESS_TOKEN] = self.ACCESS_TOKEN_ENV_VAR
        else:
            var_mapping[ACCESS_KEY_ID] = mapping.get(
                ACCESS_KEY_ID, self.ACCESS_KEY_ID_ENV_VAR)
            var_mapping[PRIVATE_KEY] = mapping.get(
                PRIVATE_KEY, self.PRIVATE_KEY_ENV_VAR)
            var_mapping[ACCESS_TOKEN] = mapping.get(
                ACCESS_TOKEN, self.ACCESS_TOKEN_ENV_VAR)
        return var_mapping

    def load(self):
        """
        Search for credentials in explicit environment variables.
        """
        if self._mapping[ACCESS_KEY_ID] in self.environ:
            access_key_id, private_key = self._extract_creds_from_mapping(
                self.environ, self._mapping[ACCESS_KEY_ID],
                self._mapping[PRIVATE_KEY])
            LOG.info('Found credentials in environment variables.')
            private_key_value = None
            # For compatibility, assume the PRIVATE_KEY is a path to a file
            # containing the key. Only if there is no file, should the value
            # be checked to see if it's an actual key.
            if not os.path.isfile(private_key):
                if Ed25519v1Auth.detect_private_key(private_key):
                    private_key_value = private_key
                else:
                    raise NoCredentialsError(
                        err_msg='Private key file {} does not exist'.format(private_key))
            else:
                private_key_value = open(private_key).read()
            return Credentials(access_key_id=access_key_id,
                               private_key=private_key_value,
                               method=self.METHOD)
        elif self._mapping[ACCESS_TOKEN] in self.environ:
            access_token = self._extract_creds_from_mapping(
                self.environ, self._mapping[ACCESS_TOKEN])
            LOG.info('Found access token in environment variables.')
            if not AccessTokenAuth.is_access_token(access_token):
                LOG.debug('Invalid access token {}'.format(access_token))
                raise NoCredentialsError(
                    err_msg='Invalid access token (see debug log for value)')
            return Credentials(access_token=access_token,
                               method=self.METHOD)
        else:
            return None


class CredentialResolver(object):

    def __init__(self, providers):
        self.providers = providers

    def insert_before(self, name, credential_provider):
        """
        Inserts a new instance of ``CredentialProvider`` into the chain that will
        be tried before an existing one.
        """
        try:
            offset = [p.METHOD for p in self.providers].index(name)
        except ValueError:
            raise UnknownCredentialError(name=name)
        self.providers.insert(offset, credential_provider)

    def insert_after(self, name, credential_provider):
        """
        Inserts a new type of ``Credentials`` instance into the chain that will
        be tried after an existing one.
        """
        offset = self._get_provider_offset(name)
        self.providers.insert(offset + 1, credential_provider)

    def remove(self, name):
        """
        Removes a given ``Credentials`` instance from the chain.
        """
        available_methods = [p.METHOD for p in self.providers]
        if name not in available_methods:
            # It's not present. Fail silently.
            return

        offset = available_methods.index(name)
        self.providers.pop(offset)

    def get_provider(self, name):
        """
        Return a credential provider by name.
        """
        return self.providers[self._get_provider_offset(name)]

    def _get_provider_offset(self, name):
        try:
            return [p.METHOD for p in self.providers].index(name)
        except ValueError:
            raise UnknownCredentialError(name=name)

    def load_credentials(self):
        """
        Goes through the credentials chain, returning the first ``Credentials``
        that could be loaded.
        """

        # Grab this during the scan in case no credentials are available.
        creds_expanded_path = None

        # First provider to return a non-None response wins.
        for provider in self.providers:
            LOG.debug("Looking for credentials via: %s", provider.METHOD)
            if isinstance(provider, SharedCredentialProvider):
                creds_expanded_path = provider.get_creds_expanded_path()
            creds = provider.load()
            if creds is not None:
                return creds

        err_msg = "No credentials found anywhere in chain"
        if creds_expanded_path:
            err_msg += ". The shared credentials file should be stored at {}"\
                .format(creds_expanded_path)

        raise NoCredentialsError(err_msg=err_msg)


class AuthConfigFile(CredentialProvider):
    METHOD = 'auth_config_file'

    def __init__(self, conf):
        super(AuthConfigFile, self).__init__()
        self._conf = conf

    def load(self):
        """
        load the credential from the json configuration file.
        """
        if self._conf is None:
            return None

        if not os.path.isfile(self._conf):
            raise NoCredentialsError(
                err_msg="Config file {} not found".format(self._conf))
        try:
            conf = json.loads(open(self._conf).read())
        except Exception:
            LOG.debug("Could not read conf: %s", exc_info=True)
            return None

        if ACCESS_KEY_ID in conf or PRIVATE_KEY in conf:
            LOG.debug('Found credentials for key: %s in configuration file.',
                      conf[ACCESS_KEY_ID])
            access_key_id, private_key = self._extract_creds_from_mapping(
                conf,
                ACCESS_KEY_ID,
                PRIVATE_KEY)
            return Credentials(access_key_id=access_key_id,
                               private_key=private_key,
                               method=self.METHOD)
        elif ACCESS_TOKEN in conf:
            LOG.debug('Found access-token in configuration file.')
            access_token = self._extract_creds_from_mapping(conf, ACCESS_TOKEN)
            return Credentials(access_token=access_token,
                               method=self.METHOD)
        else:
            cred_vars = '%s or %s' % (ACCESS_KEY_ID, ACCESS_TOKEN)
            LOG.debug('Auth config file is missing required key %s',
                      cred_vars)
            raise MalformedCredentialsError(provider=self.METHOD,
                                            cred_var=cred_vars)


class SharedCredentialProvider(CredentialProvider):
    METHOD = 'shared-credentials-file'

    def __init__(self, creds_filename, profile_name):
        self._creds_filename = creds_filename
        self._creds_expanded_path = os.path.expanduser(creds_filename)
        self._profile_name = profile_name

    def get_creds_expanded_path(self):
        return self._creds_expanded_path

    def load(self):
        try:
            available_creds = raw_config_parse(self._creds_filename)
        except ConfigNotFound:
            LOG.debug("Shared credentials file {} not found".format(self._creds_filename))
            return None
        if self._profile_name in available_creds:
            config = available_creds[self._profile_name]

            if CDP_ACCESS_KEY_ID_KEY_NAME in config or \
                    CDP_PRIVATE_KEY_KEY_NAME in config:
                access_key_id, private_key = self._extract_creds_from_mapping(
                    config, CDP_ACCESS_KEY_ID_KEY_NAME, CDP_PRIVATE_KEY_KEY_NAME)
                # We store the private key in the credentials file as a one-line
                # value in which the newlines in the PEM file are replaced with
                # '\n'. We need to replace them back as the RawConfigParser we use
                # does not do it for us. Note that if the value in the configuration
                # IS a PEM formatted value this is a no-op.
                private_key = private_key.replace('\\n', '\n')
                LOG.info("Found credentials in shared credentials file: %s",
                         self._creds_filename)
                return Credentials(access_key_id=access_key_id,
                                   private_key=private_key,
                                   method=self.METHOD)
            elif CDP_ACCESS_TOKEN_KEY_NAME in config:
                access_token = self._extract_creds_from_mapping(
                    config, CDP_ACCESS_TOKEN_KEY_NAME)
                LOG.info("Found access-token in shared credentials file: %s",
                         self._creds_filename)
                return Credentials(access_token=access_token,
                                   method=self.METHOD)
            else:
                return None


class ParamsProvider(CredentialProvider):
    """
    Support for access-token parameter only.
    """

    METHOD = 'params'

    def __init__(self, params):
        if params is None:
            self._access_token = None
        else:
            self._access_token = getattr(params, 'access_token', None)

    def load(self):
        if bool(self._access_token):
            return Credentials(access_token=self._access_token,
                               method=self.METHOD)
        return None



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import glob
import hashlib
import os
import pickle

from cdpcli import CDPCLI_ROOT
from cdpcli.compat import json
from cdpcli.compat import OrderedDict
from cdpcli.exceptions import DataNotFoundError
from cdpcli.exceptions import ValidationError
import yaml

DEFAULT_CACHE_DIR = os.path.join(os.path.expanduser("~"), ".cdp", "model-cache")


def instance_cache(func):
    """Cache the result of a method on a per instance basis.

    This is not a general purpose caching decorator. In order for this to be used,
    it must be used on methods on an instance, and that instance *must* provide a
    ``self._cache`` dictionary.

    """
    def _wrapper(self, *args, **kwargs):
        key = (func.__name__,) + args
        for pair in sorted(kwargs.items()):
            key += pair
        if key in self._cache:
            return self._cache[key]
        data = func(self, *args, **kwargs)
        self._cache[key] = data
        return data
    return _wrapper


class JSONFileLoader(object):
    def exists(self, file_path):
        return os.path.isfile(file_path)

    def load_file(self, file_path):
        if not os.path.isfile(file_path):
            return

        # By default the file will be opened with locale encoding on Python 3.
        # We specify "utf8" here to ensure the correct behavior.
        with open(file_path, 'rb') as fp:
            payload = fp.read().decode('utf-8')
            return json.loads(payload, object_pairs_hook=OrderedDict)


class YAMLFileLoader(object):
    SERVICE_ALIAS_KEY = 'x-service-alias'
    CACHE_BASE_DIR = os.environ.get("CDP_MODEL_CACHE_DIR", DEFAULT_CACHE_DIR)

    def exists(self, file_path):
        return os.path.isfile(file_path)

    def load_file(self, file_path):
        if not os.path.isfile(file_path):
            return
        # By default the file will be opened with locale encoding on Python 3.
        # We open the file in binary mode and then decode it specifically as 'utf8'
        # below to ensure the correct behavior.
        with open(file_path, 'rb') as fp:
            payload_bytes = fp.read()
        return self._pickle_cache(
            hashlib.sha1(payload_bytes).hexdigest(),
            lambda: self._ordered_load(payload_bytes.decode('utf-8')))

    def _pickle_cache(self, cache_key, func):
        """
        Cache the results of 'func' by storing the result pickled on disk.
        'cache_key' should be a hash that uniquely identifies the invocation.
        """
        cache_path = os.path.join(self.CACHE_BASE_DIR, "%s.pickle" % cache_key)
        try:
            with open(cache_path, "rb") as cache:
                try:
                    return pickle.load(cache)
                except Exception:
                    # If we have a deserialization error, remove the corrupt
                    # cache.
                    os.unlink(cache_path)
        except Exception:
            # Ignore failures to read the cache (file may not exist, etc).
            pass
        ret = func()
        # Try to write back to the cache.
        try:
            if not os.path.exists(self.CACHE_BASE_DIR):
                os.makedirs(self.CACHE_BASE_DIR)
            tmp_path = cache_path + ".tmp"
            with open(tmp_path, "wb") as cache:
                pickle.dump(ret, cache)
            os.rename(tmp_path, cache_path)
        except Exception:
            # Failure to write the cache isn't fatal (maybe we are on a read-only
            # filesystem, etc).
            pass
        return ret

    def load_alias(self, file_path):
        data = self.load_file(file_path)
        if data is not None and YAMLFileLoader.SERVICE_ALIAS_KEY in data:
            return data[YAMLFileLoader.SERVICE_ALIAS_KEY]

    def _ordered_load(self, stream):
        class OrderedLoader(yaml.SafeLoader):
            pass

        def construct_mapping(loader, node):
            loader.flatten_mapping(node)
            return OrderedDict(loader.construct_pairs(node))

        OrderedLoader.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            construct_mapping)
        return yaml.load(stream, OrderedLoader)


class Loader(object):
    """Find and load data models.

    This class will handle searching for and loading data models.

    """

    # The included models in cdpcli/data/ that we ship with cdpcli
    BUILTIN_DATA_PATH = os.path.join(CDPCLI_ROOT, 'data')
    # For convenience we automatically add ~/.cdp/models to the data path.
    CUSTOMER_DATA_PATH = os.path.join(os.path.expanduser('~'), '.cdp', 'models')

    def __init__(self):
        self._cache = {}
        # Note: the order of search paths determines the YAML file in case if file
        # names are the same (the first one wins). See self._service_directory method.
        self._search_paths = [self.CUSTOMER_DATA_PATH, self.BUILTIN_DATA_PATH]
        self._directory, self._aliases = self._service_directory()

    # returns a list of actual existing directories among search paths
    def _potential_locations(self):
        for path in self._search_paths:
            if os.path.isdir(path):
                yield path

    # lists potential service names in a given search path
    # which is obtained by shallow listing of subdirectories
    def _potential_services(self, path):
        for child in os.listdir(path):
            full_path = os.path.join(path, child)
            if os.path.isdir(full_path):
                yield (child, full_path)

    def _load_builtin_aliases(self):
        """Load aliases of built-in services generated at build time"""
        alias_path = os.path.join(self.BUILTIN_DATA_PATH, "aliases.yaml")
        with open(alias_path, 'r') as r:
            return yaml.safe_load(r)

    # build a 1:1 map of services to their YAML definition files
    # if multiple definitions are found in search path we will use one of them
    def _service_directory(self):
        dir = {}
        aliases = set()
        for possible_path in self._potential_locations():
            for (service_name, service_path) in self._potential_services(possible_path):
                for file in glob.glob(os.path.join(service_path, "*.yaml")):
                    # accumulate the last path in ascending lexical order,
                    # by filename (not full path) across all search paths
                    if service_name not in dir:
                        dir[service_name] = file
                    elif os.path.split(dir[service_name])[-1] != os.path.split(file)[-1]:
                        dir[service_name] = max(dir[service_name], file,
                                                key=lambda x: os.path.split(x)[-1])
        # Determine service aliases.
        # We expect all built-in services to have a built-in alias.
        # Customer-provided services are parsed to determine aliases.
        builtin_aliases = self._load_builtin_aliases()
        for service_name in list(dir.keys()):
            alias = None
            service_fullpath = dir[service_name]
            if service_fullpath.startswith(self.BUILTIN_DATA_PATH):
                if service_name not in builtin_aliases:
                    raise Exception("Could not find alias for service " + service_name)
                alias = builtin_aliases[service_name]
            else:
                alias = YAMLFileLoader().load_alias(dir[service_name])
            if alias is not None:
                dir[alias] = dir[service_name]
                aliases.add(alias)
        return dir, aliases

    # load the first json config file from one of search locations
    def load_json(self, name):
        loader = JSONFileLoader()
        for possible_path in self._potential_locations():
            full_path = os.path.join(possible_path, name)
            found = loader.load_file(full_path)
            if found is not None:
                return found
        raise DataNotFoundError(data_path=name)

    # returns a list of service names in directory
    def list_available_services(self):
        services = self._directory.keys()
        return sorted(services)

    # load YAML data by service name
    @instance_cache
    def load_service_data(self, service_name):
        if service_name not in self._directory.keys():
            raise ValidationError(value=service_name,
                                  param='service_name',
                                  type_name='str')
        return YAMLFileLoader().load_file(self._directory[service_name])

    # wether the input service name is a service alias
    def is_service_alias(self, service_name):
        return service_name in self._aliases



# Licensed to Cloudera, Inc. under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  Cloudera, Inc. licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy


SAMPLE_DESCRIBE_CLUSTER = {
    "cluster":
        {
            "status": "CREATED",
            "serviceType": "HIVE_ON_SPARK",
            "networkSettingsName": "net-settings-2",
            "workersGroupSize": 4,
            "clusterName": "cloudtrail-analyze-final",
            "crn": "crn:cws:dataeng:us-west-1:"
                   "id1:cluster:"
                   "cloudtrail-analyze-final/"
                   "id2",
            "cloudAccessAccountName": "caa2",
            "instanceSettingsName": "instance-settings-2",
            "creationDate": "2016-10-21T20:39:35.446000+00:00",
            "clouderaManagerEndpoint": {
                "port": 7180,
                "privateIpAddress": "162.88.88.88",
                "publicIpAddress": "10.88.88.88",

            },
            "instanceType": "c4.2xlarge",
            "cdhVersion": "CDH58"
        }
}

SAMPLE_DESCRIBE_SECURE_CLUSTER = copy.deepcopy(SAMPLE_DESCRIBE_CLUSTER)
SAMPLE_DESCRIBE_SECURE_CLUSTER["cluster"]["clouderaManagerEndpoint"]["port"] = 7183



# Copyright (c) 2012-2013 Mitch Garnaat http://garnaat.org/
# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


def build_retry_config(retry_model, definitions):
    retry_config = {'__default__': retry_model.get('__default__', {})}
    resolve_references(retry_config, definitions)
    return retry_config


def resolve_references(config, definitions):
    """Recursively replace $ref keys.

    To cut down on duplication, common definitions can be declared
    (and passed in via the ``definitions`` attribute) and then
    references as {"$ref": "name"}, when this happens the reference
    dict is placed with the value from the ``definition`` dict.

    This is recursively done.

    """
    for key, value in config.items():
        if isinstance(value, dict):
            if len(value) == 1 and list(value.keys())[0] == '$ref':
                # Then we need to resolve this reference.
                config[key] = definitions[list(value.values())[0]]
            else:
                resolve_references(value, definitions)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli import LIST_TYPE
from cdpcli import MAP_TYPE
from cdpcli import OBJECT_TYPE
from cdpcli import SCALAR_TYPES
from cdpcli import VERSION
from cdpcli import xform_name
from cdpcli.argprocess import ParamShorthandDocGen
from cdpcli.extensions.paginate import add_paging_description
from cdpcli.model import StringShape


MANUAL_SECTION = '1'
MANUAL_GROUP = 'CDP CLI'


def _is_argument_hidden(argument):
    if getattr(argument, '_UNDOCUMENTED', False):
        return True
    return False


def generate_doc(generator, help_command):
    generator.doc_title(help_command)
    generator.doc_docinfo(help_command)
    generator.doc_description(help_command)
    generator.doc_synopsis_start(help_command)
    if help_command.arg_table:
        for arg_name in help_command.arg_table:
            generator.doc_synopsis_option(arg_name, help_command)
    generator.doc_synopsis_end(help_command)
    generator.doc_options_start(help_command)
    if help_command.arg_table:
        for arg_name in help_command.arg_table:
            if _is_argument_hidden(help_command.arg_table[arg_name]):
                continue
            generator.doc_option_start(arg_name, help_command)
            generator.doc_option(arg_name, help_command)
            generator.doc_option_example(arg_name, help_command)
            generator.doc_option_form_factors(arg_name, help_command)
            generator.doc_option_end(arg_name, help_command)
    generator.doc_options_end(help_command)
    generator.doc_subitems_start(help_command)
    if help_command.command_table:
        for command_name in sorted(help_command.command_table.keys()):
            if hasattr(help_command.command_table[command_name],
                       '_UNDOCUMENTED'):
                continue
            generator.doc_subitem(command_name, help_command)
    generator.doc_subitems_end(help_command)
    generator.doc_examples(help_command)
    generator.doc_output(help_command)
    generator.doc_form_factors(help_command)
    generator.doc_relateditems_start(help_command)
    if help_command.related_items:
        for related_item in sorted(help_command.related_items):
            generator.doc_relateditem(help_command, related_item)


class CLIDocumentGenerator(object):

    def __init__(self, help_command):
        self.help_command = help_command
        self.help_command.doc.translation_map = self.build_translation_map()
        self._arg_groups = self._build_arg_table_groups(help_command)
        self._documented_arg_groups = []

    def _build_arg_table_groups(self, help_command):
        arg_groups = {}
        for name, arg in help_command.arg_table.items():
            if arg.group_name is not None:
                arg_groups.setdefault(arg.group_name, []).append(arg)
        return arg_groups

    def build_translation_map(self):
        return dict()

    # These are default doc handlers that apply in the general case.

    def doc_title(self, help_command):
        doc = help_command.doc
        doc.style.new_paragraph()
        reference = help_command.command_lineage.replace('.', ' ')
        if reference != 'cdp':
            reference = 'cdp ' + reference
        doc.writeln('.. _cli:%s:' % reference)
        doc.style.h1(help_command.name)

    def doc_docinfo(self, help_command):
        pass

    def doc_description(self, help_command):
        doc = help_command.doc
        doc.style.h2('Description')
        doc.include_doc_string(help_command.description)
        doc.style.new_paragraph()

    def doc_synopsis_start(self, help_command):
        self._documented_arg_groups = []
        doc = help_command.doc
        doc.style.h2('Synopsis')
        doc.style.start_codeblock()
        doc.writeln('%s' % help_command.name)

    def doc_synopsis_option(self, arg_name, help_command):
        doc = help_command.doc
        argument = help_command.arg_table[arg_name]
        if _is_argument_hidden(argument):
            return
        if argument.group_name in self._arg_groups:
            if argument.group_name in self._documented_arg_groups:
                # This arg is already documented so we can move on.
                return
            option_str = ' | '.join(
                [a.cli_name for a in
                 self._arg_groups[argument.group_name]])
            self._documented_arg_groups.append(argument.group_name)
        elif argument.cli_type_name == 'blob':
            option_str = '%s <blob>' % argument.cli_name
        else:
            option_str = '%s <value>' % argument.cli_name
        if not argument.required:
            option_str = '[%s]' % option_str
        doc.writeln('%s' % option_str)

    def doc_synopsis_end(self, help_command):
        doc = help_command.doc
        doc.style.end_codeblock()
        # Reset the documented arg groups for other sections
        # that may document args (the detailed docs following
        # the synopsis).
        self._documented_arg_groups = []

    def doc_options_start(self, help_command):
        pass

    def doc_option_start(self, arg_name, help_command):
        pass

    def doc_option(self, arg_name, help_command):
        pass

    def doc_option_example(self, arg_name, help_command):
        pass

    def doc_option_form_factors(self, arg_name, help_command):
        pass

    def doc_option_end(self, arg_name, help_command):
        pass

    def doc_options_end(self, help_command):
        pass

    def doc_relateditems_start(self, help_command):
        if help_command.related_items:
            doc = help_command.doc
            doc.style.h2('See Also')

    def doc_relateditem(self, help_command, related_item):
        doc = help_command.doc
        doc.write('* ')
        doc.style.sphinx_reference_label(
            label='cli:%s' % related_item,
            text=related_item
        )
        doc.write('\n')

    def doc_subitems_start(self, help_command):
        pass

    def doc_subitems_end(self, help_command):
        pass

    def doc_examples(self, help_command):
        pass

    def doc_output(self, help_command):
        pass

    def doc_form_factors(self, help_command):
        pass


class ProviderDocumentGenerator(CLIDocumentGenerator):

    def doc_synopsis_start(self, help_command):
        doc = help_command.doc
        doc.style.h2('Synopsis')
        doc.style.codeblock(help_command.synopsis)
        doc.include_doc_string(help_command.help_usage)

    def doc_synopsis_option(self, arg_name, help_command):
        pass

    def doc_synopsis_end(self, help_command):
        doc = help_command.doc
        doc.style.new_paragraph()

    def doc_options_start(self, help_command):
        doc = help_command.doc
        doc.style.h2('Options')

    def doc_option_start(self, arg_name, help_command):
        pass

    def doc_option(self, arg_name, help_command):
        doc = help_command.doc
        argument = help_command.arg_table[arg_name]
        doc.writeln('``%s`` (%s)' % (argument.cli_name,
                                     argument.cli_type_name))
        doc.include_doc_string(argument.documentation)
        if argument.choices:
            doc.style.start_ul()
            for choice in argument.choices:
                doc.style.li(choice)
            doc.style.end_ul()

    def doc_option_example(self, arg_name, help_command):
        pass

    def doc_option_form_factors(self, arg_name, help_command):
        pass

    def doc_option_end(self, arg_name, help_command):
        pass

    def doc_options_end(self, help_command):
        pass

    def doc_subitems_start(self, help_command):
        doc = help_command.doc
        doc.style.h2('Available Commands')
        doc.style.toctree()

    def doc_subitem(self, command_name, help_command):
        doc = help_command.doc
        file_name = '%s/index' % command_name
        doc.style.tocitem(command_name, file_name=file_name)


class ServiceDocumentGenerator(CLIDocumentGenerator):

    def build_translation_map(self):
        d = {}
        service_model = self.help_command.obj
        for operation_name in service_model.operation_names:
            d[operation_name] = xform_name(operation_name, '-')
        return d

    # A service document has no synopsis.
    def doc_synopsis_start(self, help_command):
        pass

    def doc_synopsis_option(self, arg_name, help_command):
        pass

    def doc_synopsis_end(self, help_command):
        pass

    # A service document has no option section.
    def doc_options_start(self, help_command):
        pass

    def doc_option_start(self, arg_name, help_command):
        pass

    def doc_option(self, arg_name, help_command):
        pass

    def doc_option_example(self, arg_name, help_command):
        pass

    def doc_option_form_factors(self, arg_name, help_command):
        pass

    def doc_option_end(self, arg_name, help_command):
        pass

    def doc_options_end(self, help_command):
        pass

    def doc_docinfo(self, help_command):
        doc = help_command.doc
        service_model = help_command.obj
        doc.style.simple_field('subtitle', service_model.service_title)
        service_version = service_model.service_version
        # Goes away once each service has its own version number
        if service_version == '__API_VERSION__':
            service_version = VERSION
        doc.style.simple_field('version', service_version)
        if help_command.include_man_fields:
            doc.style.simple_field('manual_section', MANUAL_SECTION)
            doc.style.simple_field('manual_group', MANUAL_GROUP)

    def doc_description(self, help_command):
        doc = help_command.doc
        service_model = help_command.obj
        doc.style.h2('Description')
        # TODO: need a documentation attribute.
        doc.include_doc_string(service_model.documentation)

    def doc_subitems_start(self, help_command):
        doc = help_command.doc
        doc.style.h2('Available Subcommands')
        doc.style.toctree()

    def doc_subitem(self, command_name, help_command):
        doc = help_command.doc
        subcommand = help_command.command_table[command_name]
        subcommand_table = getattr(subcommand, 'subcommand_table', {})
        # If the subcommand table has commands in it,
        # direct the subitem to the command's index because
        # it has more subcommands to be documented.
        if (len(subcommand_table) > 0):
            file_name = '%s/index' % command_name
            doc.style.tocitem(command_name, file_name=file_name)
        else:
            doc.style.tocitem(command_name)


class OperationDocumentGenerator(CLIDocumentGenerator):

    def build_translation_map(self):
        operation_model = self.help_command.obj
        d = {}
        for cli_name, cli_argument in self.help_command.arg_table.items():
            if cli_argument.argument_model is not None:
                argument_name = cli_argument.argument_model.name
                if argument_name in d:
                    previous_mapping = d[argument_name]
                    # If the argument name is a boolean argument, we want the
                    # the translation to default to the one that does not start
                    # with --no-. So we check if the cli parameter currently
                    # being used starts with no- and if stripping off the no-
                    # results in the new proposed cli argument name. If it
                    # does, we assume we have the postive form of the argument
                    # which is the name we want to use in doc translations.
                    if cli_argument.cli_type_name == 'boolean' and \
                       previous_mapping.startswith('no-') and \
                       cli_name == previous_mapping[3:]:
                        d[argument_name] = cli_name
                else:
                    d[argument_name] = cli_name
        for operation_name in operation_model.service_model.operation_names:
            d[operation_name] = xform_name(operation_name, '-')
        return d

    def doc_docinfo(self, help_command):
        doc = help_command.doc
        operation_model = help_command.obj
        # Internal commands like "configure" may lack these (see BasicDocHandler)
        if hasattr(operation_model, 'summary'):
            doc.style.simple_field('subtitle', operation_model.summary)
        if hasattr(operation_model, 'service_model') and\
                hasattr(operation_model.service_model, 'service_version'):
            service_version = operation_model.service_model.service_version
            # Goes away once each service has its own version number
            if service_version == '__API_VERSION__':
                service_version = VERSION
            doc.style.simple_field('version', service_version)
        if help_command.include_man_fields:
            doc.style.simple_field('manual_section', MANUAL_SECTION)
            doc.style.simple_field('manual_group', MANUAL_GROUP)

    def doc_description(self, help_command):
        doc = help_command.doc
        operation_model = help_command.obj
        doc.style.h2('Description')
        doc.include_doc_string(operation_model.documentation)
        add_paging_description(help_command)

    def _json_example_value_name(self, argument_model, include_enum_values=True):
        # If include_enum_values is True, then the valid enum values
        # are included as the sample JSON value.
        if isinstance(argument_model, StringShape):
            if argument_model.enum and include_enum_values:
                choices = argument_model.enum
                return '|'.join(['"%s"' % c for c in choices])
            elif argument_model.supported_options and include_enum_values:
                choices = argument_model.supported_options
                return '|'.join(['"%s"' % c for c in choices])
            else:
                return '"string"'
        elif argument_model.type_name == 'boolean':
            return 'true|false'
        else:
            return '%s' % argument_model.type_name

    def _json_example(self, doc, argument_model, stack):
        if argument_model.name in stack:
            # Document the recursion once, otherwise just
            # note the fact that it's recursive and return.
            if stack.count(argument_model.name) > 1:
                if argument_model.type_name == OBJECT_TYPE:
                    doc.write('{ ... recursive ... }')
                return
        stack.append(argument_model.name)
        try:
            self._do_json_example(doc, argument_model, stack)
        finally:
            stack.pop()

    def _do_json_example(self, doc, argument_model, stack):
        if argument_model.type_name == LIST_TYPE:
            doc.write('[')
            if argument_model.member.type_name in SCALAR_TYPES:
                doc.write('%s, ...' %
                          self._json_example_value_name(argument_model.member))
            else:
                doc.style.indent()
                doc.style.new_line()
                self._json_example(doc, argument_model.member, stack)
                doc.style.new_line()
                doc.write('...')
                doc.style.dedent()
                doc.style.new_line()
            doc.write(']')
        elif argument_model.type_name == OBJECT_TYPE:
            doc.write('{')
            doc.style.indent()
            doc.style.new_line()
            self._doc_input_structure_members(doc, argument_model, stack)
            doc.style.dedent()
            doc.style.new_line()
            doc.write('}')
        elif argument_model.type_name == MAP_TYPE:
            doc.write('{')
            doc.style.indent()
            key_string = self._json_example_value_name(argument_model.key)
            doc.write('%s: ' % key_string)
            if argument_model.value.type_name in SCALAR_TYPES:
                doc.write(self._json_example_value_name(argument_model.value))
            else:
                doc.style.indent()
                self._json_example(doc, argument_model.value, stack)
                doc.style.dedent()
            doc.style.new_line()
            doc.write('...')
            doc.style.dedent()
            doc.write('}')

    def _doc_input_structure_members(self, doc, argument_model, stack):
        need_comma = False
        members = argument_model.members
        for i, member_name in enumerate(members):
            member_model = members[member_name]
            member_type_name = member_model.type_name
            if member_type_name in SCALAR_TYPES:
                if need_comma:
                    doc.write(',')
                    doc.style.new_line()
                doc.write('"%s": %s' % (member_name,
                                        self._json_example_value_name(member_model)))
                need_comma = True
            elif member_type_name == OBJECT_TYPE:
                if need_comma:
                    doc.write(',')
                    doc.style.new_line()
                doc.write('"%s": ' % member_name)
                self._json_example(doc, member_model, stack)
                need_comma = True
            elif member_type_name == MAP_TYPE:
                if need_comma:
                    doc.write(',')
                    doc.style.new_line()
                doc.write('"%s": ' % member_name)
                self._json_example(doc, member_model, stack)
                need_comma = True
            elif member_type_name == LIST_TYPE:
                if need_comma:
                    doc.write(',')
                    doc.style.new_line()
                doc.write('"%s": ' % member_name)
                self._json_example(doc, member_model, stack)
                need_comma = True

    def doc_options_start(self, help_command):
        doc = help_command.doc
        doc.style.h2('Options')
        if not help_command.arg_table:
            doc.write('*None*\n')

    def doc_option_start(self, arg_name, help_command):
        pass

    def doc_option(self, arg_name, help_command):
        doc = help_command.doc
        argument = help_command.arg_table[arg_name]
        if _is_argument_hidden(argument):
            return
        if argument.group_name in self._arg_groups:
            if argument.group_name in self._documented_arg_groups:
                # This arg is already documented so we can move on.
                return
            name = ' | '.join(
                ['``%s``' % a.cli_name for a in
                 self._arg_groups[argument.group_name]])
        else:
            name = '``%s``' % argument.cli_name
        doc.write('%s (%s)\n' % (name, argument.cli_type_name))
        doc.style.indent()
        doc.style.new_paragraph()
        doc.include_doc_string(argument.documentation)
        self._document_enums(argument, doc)
        if hasattr(argument.argument_model, 'members'):
            doc.style.new_paragraph()
            for member_name, member_shape in argument.argument_model.members.items():
                self.doc_member(doc, member_name, member_shape, stack=[])
        doc.style.dedent()
        doc.style.new_paragraph()

    def doc_option_example(self, arg_name, help_command):
        doc = help_command.doc
        cli_argument = help_command.arg_table[arg_name]
        if _is_argument_hidden(cli_argument):
            return
        if cli_argument.group_name in self._arg_groups:
            if cli_argument.group_name in self._documented_arg_groups:
                # Args with group_names (boolean args) don't
                # need to generate example syntax.
                return
        argument_model = cli_argument.argument_model
        docgen = ParamShorthandDocGen()
        if docgen.supports_shorthand(cli_argument.argument_model):
            example_shorthand_syntax = docgen.generate_shorthand_example(
                cli_argument.cli_name, cli_argument.argument_model)
            if example_shorthand_syntax is None:
                # If the shorthand syntax returns a value of None,
                # this indicates to us that there is no example
                # needed for this param so we can immediately
                # return.
                return
            if example_shorthand_syntax:
                doc.style.new_paragraph()
                doc.write('Shorthand Syntax')
                doc.style.start_codeblock()
                for example_line in example_shorthand_syntax.splitlines():
                    doc.writeln(example_line)
                doc.style.end_codeblock()
        if argument_model is not None and \
           argument_model.type_name == LIST_TYPE and \
           argument_model.member.type_name in SCALAR_TYPES:
            # A list of scalars is special.  While you *can* use
            # JSON ( ["foo", "bar", "baz"] ), you can also just
            # use the argparse behavior of space separated lists.
            # "foo" "bar" "baz".  In fact we don't even want to
            # document the JSON syntax in this case.
            member = argument_model.member
            doc.style.new_paragraph()
            doc.write('Syntax')
            doc.style.start_codeblock()
            example_type = self._json_example_value_name(
                member, include_enum_values=False)
            doc.write('%s %s ...' % (example_type, example_type))
            if isinstance(member, StringShape):
                if member.enum:
                    self._write_possible_values(doc, member.enum)
                if member.supported_options:
                    self._write_possible_values(doc, member.supported_options)
            doc.style.end_codeblock()
            doc.style.new_paragraph()
        elif cli_argument.cli_type_name not in SCALAR_TYPES:
            doc.style.new_paragraph()
            doc.write('JSON Syntax')
            doc.style.start_codeblock()
            self._json_example(doc, argument_model, stack=[])
            doc.style.end_codeblock()
            doc.style.new_paragraph()

    def doc_option_form_factors(self, arg_name, help_command):
        doc = help_command.doc
        argument = help_command.arg_table[arg_name]
        if _is_argument_hidden(argument):
            return
        if argument.group_name in self._arg_groups:
            if argument.group_name in self._documented_arg_groups:
                # This arg is already documented so we can move on.
                return
        if argument.argument_model.form_factors is not None:
            self._write_form_factors(doc, argument.argument_model.form_factors)

    def doc_option_end(self, arg_name, help_command):
        argument = help_command.arg_table[arg_name]
        if argument.group_name in self._arg_groups:
            if argument.group_name in self._documented_arg_groups:
                # This arg is already documented so we can move on.
                return
            self._documented_arg_groups.append(argument.group_name)

    def doc_options_end(self, help_command):
        pass

    def doc_output(self, help_command):
        doc = help_command.doc
        doc.style.h2('Output')
        operation_model = help_command.obj
        output_shape = operation_model.output_shape
        if output_shape is None:
            doc.write('None')
        else:
            for member_name, member_shape in output_shape.members.items():
                self.doc_member(doc, member_name, member_shape, stack=[])

    def doc_examples(self, help_command):
        operation = help_command.command_lineage.replace('.', os.path.sep)
        doc_path = os.path.dirname(os.path.abspath(__file__))
        doc_path = os.path.join(doc_path, 'examples')
        doc_path = os.path.join(doc_path, operation) + '.rst'
        if os.path.isfile(doc_path):
            help_command.doc.style.h2('Examples')
            fp = open(doc_path)
            for line in fp.readlines():
                help_command.doc.write(line)

    def doc_form_factors(self, help_command):
        operation_model = help_command.obj
        form_factors = getattr(operation_model, 'form_factors', None)
        if not form_factors:
            service_model = getattr(operation_model, 'service_model', None)
            if service_model:
                form_factors = getattr(service_model, 'form_factors', None)
        if form_factors:
            doc = help_command.doc
            doc.style.h2('Form Factors')
            doc.write(', '.join(form_factors))
            doc.style.new_paragraph()

    def doc_member(self, doc, member_name, member_shape, stack):
        if member_shape.name in stack:
            # Document the recursion once, otherwise just
            # note the fact that it's recursive and return.
            if stack.count(member_shape.name) > 1:
                if member_shape.type_name == OBJECT_TYPE:
                    doc.write('( ... recursive ... )')
                return
        stack.append(member_shape.name)
        try:
            self._do_doc_member(doc, member_name,
                                member_shape, stack)
        finally:
            stack.pop()

    def _do_doc_member(self, doc, member_name, member_shape, stack):
        docs = member_shape.documentation
        if member_name:
            doc.write('%s -> (%s)' % (member_name, member_shape.type_name))
        else:
            doc.write('(%s)' % member_shape.type_name)
        doc.style.indent()
        doc.style.new_paragraph()
        doc.include_doc_string(docs)
        if member_shape.form_factors is not None:
            self._write_form_factors(doc, member_shape.form_factors)
        doc.style.new_paragraph()
        member_type_name = member_shape.type_name
        if member_type_name == OBJECT_TYPE:
            for sub_name, sub_shape in member_shape.members.items():
                self.doc_member(doc, sub_name, sub_shape, stack)
        elif member_type_name == LIST_TYPE:
            self.doc_member(doc, 'item', member_shape.member, stack)
        elif member_type_name == MAP_TYPE:
            self.doc_member(doc, 'key', member_shape.key, stack)
            self.doc_member(doc, 'value', member_shape.value, stack)
        doc.style.dedent()
        doc.style.new_paragraph()

    def _document_enums(self, argument, doc):
        if hasattr(argument, 'argument_model'):
            model = argument.argument_model
            if isinstance(model, StringShape):
                if model.enum:
                    self._write_possible_values(doc, model.enum)
                if model.supported_options:
                    self._write_possible_values(doc, model.supported_options)

    def _write_possible_values(self, doc, possible_values):
        doc.style.new_paragraph()
        doc.write('Possible values:')
        doc.style.start_ul()
        for value in possible_values:
            doc.style.li('``%s``' % value)
        doc.style.end_ul()

    def _write_form_factors(self, doc, form_factors):
        doc.style.new_paragraph()
        doc.write('Form Factors: ')
        doc.write(', '.join(form_factors))
        doc.style.new_paragraph()



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from codecs import open
from os import path


def read_long_description():
    """
    Reads the long description of the package from the project README.md file.
    """
    here = path.abspath(path.dirname(__file__))

    # Get the long description from the README file
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        return f.read()


def get_requirements():
    """
    Gets the prerequisite requirements for the package.
    """
    return ["python-dateutil>=2.1,<3.0.0",
            "docutils==0.14",
            "pyyaml>=3.11",
            "colorama>=0.2.5,!=0.3.8,!=0.3.9,!=0.4.2",
            "asn1crypto>=0.21.1",
            "rsa>=3.4.2",
            "pure25519>=0.0.1",
            "ecdsa>=0.17.0",
            "requests>=2.21.0",
            "urllib3>=1.21.1"]


def get_classifiers(release):
    """
    Gets the classifiers for the package.
    """
    if release == 'public':
        classifiers = [
            'Development Status :: 5 - Production/Stable',
        ]
    elif release == 'beta':
        classifiers = [
            'Development Status :: 4 - Beta'
        ]
    else:
        classifiers = [
            'Development Status :: 3 - Alpha'
        ]

    classifiers.extend([
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ])
    return classifiers


def get_entry_points():
    """
    Gets the entry points for the package.
    """
    return {
        'console_scripts': [
            'cdp=cdpcli.clidriver:main',
            'cdp_completer=cdpcli.completer:main'
        ],
    }



# Copyright 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
from urllib.parse import urljoin

from cdpcli.exceptions import WorkloadServiceDiscoveryError


LOG = logging.getLogger('cdpcli.extensions.workload')


def register(operation_callers, operation_model):
    """
    Register an extension to run before or after the CLI command.
    To replace the original CLI caller:
    * operation_callers.insert(0, ReplacementCaller())
    * return False by the ReplacementCaller.invoke(...)
    """
    operation_callers.insert(0, WorkloadServiceDiscovery())


def set_workload_access_token(iam_client,
                              parsed_globals,
                              workload_name,
                              environment_crn):
    """
    Set global Access Token and Endpoint URL for Workload and Environment CRN
    """
    req_params = {
        'workloadName': workload_name,
        'environmentCrn': environment_crn
    }
    response = iam_client.generate_workload_auth_token(**req_params)

    workload_url = response.get('endpointUrl', None)
    workload_access_token = response.get('token', None)
    LOG.debug('Workload service-discovery succeeded. '
              'endpoint_url=%s, access_token=%s...',
              workload_url,
              workload_access_token[0:16] if workload_access_token else None)

    # remove path from the workload URL
    if workload_url is not None:
        workload_url = urljoin(workload_url, '/')
    # add 'Bearer' prefix to access token
    if workload_access_token is not None and \
       not workload_access_token.startswith('Bearer '):
        workload_access_token = 'Bearer ' + workload_access_token

    # Set the workload URL and access-token in parsed_globals, so the
    # next CLI operation caller will use them to make the connection to workload.
    parsed_globals.endpoint_url = workload_url
    parsed_globals.access_token = workload_access_token


class WorkloadServiceDiscovery(object):
    def invoke(self,
               client_creator,
               operation_model,
               parameters,
               parsed_args,
               parsed_globals):
        if parsed_globals.access_token:
            # access-token was explicitly set by user,
            # skip the workload service-discovery
            LOG.debug('Skip workload service-discovery')
            return True

        LOG.debug('Run workload service-discovery')
        service_name = operation_model.service_model.service_name
        if not service_name:
            raise WorkloadServiceDiscoveryError(err_msg='Missing service name')

        if service_name == 'dfworkload':
            workload_name = 'DF'
            environment_crn = parameters.get('environmentCrn', None)
            if not environment_crn:
                raise WorkloadServiceDiscoveryError(err_msg='Missing environment CRN')
        else:
            raise WorkloadServiceDiscoveryError(
                err_msg='Unknown service name \'%s\'' % service_name)

        iam_client = client_creator('iam')
        set_workload_access_token(iam_client, parsed_globals, workload_name,
                                  environment_crn)

        # The command processing is not finished, continue with other operation callers.
        return True



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys

from cdpcli import CDP_ACCESS_KEY_ID_KEY_NAME,\
    CDP_ENDPOINT_URL_KEY_NAME,\
    CDP_PRIVATE_KEY_KEY_NAME,\
    CDP_REGION_KEY_NAME,\
    ENDPOINT_URL_KEY_NAME,\
    FORM_FACTOR_KEY_NAME
from cdpcli.extensions.commands import BasicCommand

from . import ConfigValue, NOT_SET


class ConfigureListCommand(BasicCommand):
    NAME = 'list'
    DESCRIPTION = (
        'List CDP CLI configuration data. Each configuration item\'s name, '
        'value, source type, and source location are listed. For example, if '
        'you provide the CDP access key in an environment variable, this '
        'command will list the access key value and its source as the '
        'environment.\n'
    )
    SYNOPSIS = 'cdp configure list [--profile profile-name]'
    EXAMPLES = (
        'To show your current configuration values::\n'
        '\n'
        '  $ cdp configure list\n'
        '                Name                Value'
        '              Source Type    Source\n'
        '                ----                -----'
        '              -----------    ------\n'
        '             profile            <not set>'
        '                     None    None\n'
        '   cdp_access_key_id ****************ABCD  shared-credentials-file\n'
        '     cdp_private_key ****************EFGH  shared-credentials-file\n'
        '    cdp_endpoint_url            <not set>'
        '                     None    None\n'
        '        endpoint_url            <not set>'
        '                     None    None\n'
        '         form_factor            <not set>'
        '                     None    None\n'
        '\n'
    )

    def __init__(self, stream=sys.stdout):
        super(ConfigureListCommand, self).__init__()
        self._stream = stream

    def _run_main(self, client_creator, args, parsed_globals):
        context = client_creator.context

        self._display_config_value(ConfigValue('Value', 'Source Type', 'Source'),
                                   'Name')
        self._display_config_value(ConfigValue('-----', '-----------', '------'),
                                   '----')

        if context.profile is not None:
            profile = ConfigValue(context.profile, 'manual', '--profile')
        else:
            profile = self._lookup_config(context, 'profile')
        self._display_config_value(profile, 'profile')

        access_key, private_key = self._lookup_credentials(context)
        self._display_config_value(access_key, CDP_ACCESS_KEY_ID_KEY_NAME)
        self._display_config_value(private_key, CDP_PRIVATE_KEY_KEY_NAME)

        self._display_config_value(self._lookup_config(context, CDP_REGION_KEY_NAME),
                                   CDP_REGION_KEY_NAME)
        self._display_config_value(self._lookup_config(context,
                                                       CDP_ENDPOINT_URL_KEY_NAME),
                                   CDP_ENDPOINT_URL_KEY_NAME)
        self._display_config_value(self._lookup_config(context, ENDPOINT_URL_KEY_NAME),
                                   ENDPOINT_URL_KEY_NAME)
        self._display_config_value(self._lookup_config(context, FORM_FACTOR_KEY_NAME),
                                   FORM_FACTOR_KEY_NAME)

    def _display_config_value(self, config_value, config_name):
        truncated_value = (config_value.value[:27] + '...')\
            if len(config_value.value) > 30 else config_value.value
        self._stream.write('%20s %30s %24s    %s\n' % (
            config_name, truncated_value, config_value.source_type,
            config_value.source))

    def _lookup_credentials(self, context):
        # First try it with _lookup_config.  It's possible
        # that we don't find credentials this way.
        access_key = self._lookup_config(context, CDP_ACCESS_KEY_ID_KEY_NAME)
        if access_key.value is not NOT_SET:
            private_key = self._lookup_config(context, CDP_PRIVATE_KEY_KEY_NAME)
            access_key.mask_value()
            private_key.mask_value()
            return access_key, private_key
        else:
            # Otherwise we can try to use get_credentials().
            # This includes a few more lookup locations.
            credentials = context.get_credentials()
            if credentials is None:
                no_config = ConfigValue(NOT_SET, None, None)
                return no_config, no_config
            else:
                # For the ConfigValue, we don't track down the
                # config_variable because that info is not
                # visible from cdpcli.credentials.  I think
                # the credentials.method is sufficient to show
                # where the credentials are coming from.
                access_key = ConfigValue(credentials.access_key_id,
                                         credentials.method, '')
                private_key = ConfigValue(credentials.private_key,
                                          credentials.method, '')
                access_key.mask_value()
                private_key.mask_value()
                return access_key, private_key

    def _lookup_config(self, context, name):
        # First try to look up the variable in the env.
        value = context.get_config_variable(name, methods=('env',))
        if value is not None:
            return ConfigValue(value, 'env', context.context_var_map[name][1])
        # Then try to look up the variable in the config file.
        value = context.get_config_variable(name, methods=('config',))
        if value is not None:
            return ConfigValue(value, 'config-file',
                               context.get_config_variable('config_file'))
        else:
            return ConfigValue(NOT_SET, None, None)



# Copyright (c) 2017 Cloudera, Inc. All rights reserved.

import os
import shutil
import tempfile

from cdpcli.exceptions import DataNotFoundError
from cdpcli.exceptions import ValidationError
from cdpcli.loader import JSONFileLoader
from cdpcli.loader import Loader
from cdpcli.loader import YAMLFileLoader
from mock import patch
from nose.tools import raises
from tests import unittest


LOADER_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'loader')
TEST_TMP_DIR = tempfile.mkdtemp("foo.XXXX")


# Base class with common cases across loaders. Not an actual test. Shouldn't run.
class TestFileLoader(unittest.TestCase):
    def test_load_file(self):
        data = self.file_loader.load_file(self.valid_file_path)
        self.assertIsNotNone(data)
        self.assertEqual(len(data), 4)
        self.assertTrue('test_key_1' in data)

    def test_load_file_does_not_exist_returns_none(self):
        # None is used to indicate that the loader could not find a
        # file to load.
        self.assertIsNone(self.file_loader.load_file(self.not_found_path))

    def test_file_exists_check(self):
        self.assertTrue(self.file_loader.exists(self.valid_file_path))

    def test_file_does_not_exist_returns_false(self):
        self.assertFalse(self.file_loader.exists(self.not_found_path))

    def test_file_with_non_ascii(self):
        try:
            self.assertIsNotNone(self.file_loader.load_file(self.non_ascii_file_path))
        except UnicodeDecodeError:
            self.fail('Fail to handle data file with non-ascii characters')


class TestJSONFileLoader(TestFileLoader):
    def setUp(self):
        super(TestJSONFileLoader, self).setUp()
        self.file_loader = JSONFileLoader()
        self.data_path = LOADER_DIR
        self.valid_file_path = os.path.join(self.data_path, 'foo.json')
        self.non_ascii_file_path = os.path.join(self.data_path, 'non_ascii.json')
        self.not_found_path = os.path.join(self.data_path, 'does', 'not', 'exist')


class TestYAMLFileLoader(TestFileLoader):
    def setUp(self):
        super(TestYAMLFileLoader, self).setUp()
        self.file_loader = YAMLFileLoader()
        self.file_loader.CACHE_BASE_DIR = TEST_TMP_DIR
        self.data_path = LOADER_DIR
        self.valid_file_path = os.path.join(self.data_path, 'foo.yaml')
        self.non_ascii_file_path = os.path.join(self.data_path, 'non_ascii.yaml')
        self.not_found_path = os.path.join(self.data_path, 'does', 'not', 'exist')

    def tearDown(self):
        if os.path.exists(TEST_TMP_DIR):
            shutil.rmtree(TEST_TMP_DIR)

    def test_load_alias_does_not_exist_returns_none(self):
        self.assertIsNone(self.file_loader.load_alias(self.not_found_path))

    def test_load_alias_returns_alias(self):
        self.assertEqual(self.file_loader.load_alias(self.valid_file_path), 'bar')


# helper class encapsulating fake methods for Loader tests
class FakeFileSystem(object):
    @staticmethod
    def builtin_full_path(service, file):
        return os.path.join(Loader.BUILTIN_DATA_PATH, service, file)

    @staticmethod
    def customer_full_path(service, file):
        return os.path.join(Loader.CUSTOMER_DATA_PATH, service, file)

    # fake directories in search paths
    @staticmethod
    def listdir(path):
        return {
            Loader.BUILTIN_DATA_PATH: ['service-1', 'service-2', 'service-3', 'service-4',
                                       'service-6'],
            Loader.CUSTOMER_DATA_PATH: ['service-2', 'service-3', 'service-4',
                                        'service-5']
        }[path]

    @staticmethod
    def load_builtin_aliases():
        return {
            'service-1': 'alias-1',
            'service-2': None,
            'service-3': None,
            'service-4': None,
            'service-6': None,
        }

    # fake file layout to test various model selection behaviors
    @staticmethod
    def glob(path):
        # local aliases for brevity
        builtin = FakeFileSystem.builtin_full_path
        customer = FakeFileSystem.customer_full_path

        return {
            # built-in search path
            builtin('service-1', '*.yaml'): [builtin('service-1', 'service-1-1.yaml'),
                                             builtin('service-1', 'service-1-2.yaml')],
            builtin('service-2', '*.yaml'): [builtin('service-2', 'service-2-1.yaml')],
            builtin('service-3', '*.yaml'): [builtin('service-3', 'service-3-1.yaml')],
            builtin('service-4', '*.yaml'): [builtin('service-4', 'service-4-2.yaml')],
            builtin('service-6', '*.yaml'): [],

            # customer home dir search path
            customer('service-2', '*.yaml'): [customer('service-2', 'service-2-1.yaml')],
            customer('service-3', '*.yaml'): [customer('service-3', 'service-3-1.yaml'),
                                              customer('service-3', 'service-3-2.yaml')],
            customer('service-4', '*.yaml'): [],
            customer('service-5', '*.yaml'): [customer('service-5', 'service-5-1.yaml')]
        }[path]


class TestLoader(unittest.TestCase):
    def setUp(self):
        super(TestLoader, self).setUp()

    @patch('glob.glob', side_effect=FakeFileSystem.glob)
    @patch('os.listdir', side_effect=FakeFileSystem.listdir)
    @patch('os.path.isdir', return_value=True)
    @patch('cdpcli.loader.Loader._load_builtin_aliases',
           side_effect=FakeFileSystem.load_builtin_aliases)
    def test_list_available_services(self, *args):
        loader = Loader()
        services = loader.list_available_services()

        self.assertEqual(6, len(services))
        self.assertEqual(len(set(services)), len(services))
        self.assertEqual(0, len(set(services).symmetric_difference(['alias-1',
                                                                    'service-1',
                                                                    'service-2',
                                                                    'service-3',
                                                                    'service-4',
                                                                    'service-5'])))

    @raises(Exception)
    @patch('glob.glob', side_effect=FakeFileSystem.glob)
    @patch('os.listdir', side_effect=FakeFileSystem.listdir)
    @patch('os.path.isdir', return_value=True)
    @patch('cdpcli.loader.Loader._load_builtin_aliases',
           return_value={
            "service-2": None,
            "service-3": None,
            "service-4": None,
            "service-6": None})
    def test_missing_builtin_service_alias(self, *args):
        '''Throws an exception because built-in service-1 is missing'''
        Loader()

    @patch('glob.glob', side_effect=FakeFileSystem.glob)
    @patch('os.listdir', side_effect=FakeFileSystem.listdir)
    @patch('os.path.isdir', return_value=True)
    @patch('cdpcli.loader.Loader._load_builtin_aliases',
           side_effect=FakeFileSystem.load_builtin_aliases)
    def test_load_service_data(self, *args):
        # local aliases for brevity
        builtin = FakeFileSystem.builtin_full_path
        customer = FakeFileSystem.customer_full_path

        correct = [['service-1', builtin('service-1', 'service-1-2.yaml')],
                   ['service-4', builtin('service-4', 'service-4-2.yaml')],
                   ['service-2', customer('service-2', 'service-2-1.yaml')],
                   ['service-3', customer('service-3', 'service-3-2.yaml')],
                   ['service-5', customer('service-5', 'service-5-1.yaml')],
                   ['alias-1', builtin('service-1', 'service-1-2.yaml')]]

        with patch.object(YAMLFileLoader, 'load_alias') as mock_load_alias:
            mock_load_alias.side_effect = \
                lambda path: 'alias-1' if 'service-1' in path else None
            loader = Loader()
            for service in correct:
                with patch.object(YAMLFileLoader, 'load_file') as mock_load_file:
                    mock_load_file.side_effect = lambda path: path
                    self.assertEqual(service[1], loader.load_service_data(service[0]))
                    self.assertEqual(1, mock_load_file.call_count)

        for not_found in ['service-6', 'service-foo']:
            with patch.object(YAMLFileLoader, 'load_file') as mock_load_file:
                with self.assertRaises(ValidationError):
                    loader.load_service_data(not_found)
                self.assertEqual(0, mock_load_file.call_count)

    # ensures that the results of load_service_data is repeatable and that
    # result of the loader is cached
    @patch('glob.glob', side_effect=FakeFileSystem.glob)
    @patch('os.listdir', side_effect=FakeFileSystem.listdir)
    @patch('os.path.isdir', return_value=True)
    @patch('cdpcli.loader.Loader._load_builtin_aliases',
           side_effect=FakeFileSystem.load_builtin_aliases)
    def test_load_service_data_repeatable_cached(self, *args):
        with patch.object(YAMLFileLoader, 'load_file'):
            loader = Loader()
        # warm up with both args and kwargs, ensure results are identical
        with patch.object(YAMLFileLoader, 'load_file') as mock_load_file:
            mock_load_file.side_effect = lambda path: path
            data = loader.load_service_data('service-1')
            self.assertEqual(data, loader.load_service_data(service_name='service-1'))
            self.assertEqual(2, mock_load_file.call_count)
        # ensure results are repeatable and are coming from cache (not calls to load_file)
        with patch.object(YAMLFileLoader, 'load_file') as mock_load_file:
            mock_load_file.side_effect = lambda path: path
            self.assertEqual(data, loader.load_service_data('service-1'))
            self.assertEqual(data, loader.load_service_data(service_name='service-1'))
            self.assertEqual(0, mock_load_file.call_count)

    # loading JSON from search paths by filename
    def test_load_json(self, *args):
        cases = [
            [lambda path: path == Loader.BUILTIN_DATA_PATH, Loader.BUILTIN_DATA_PATH,
             "Loader should use builtin data path."],
            [lambda path: path == Loader.CUSTOMER_DATA_PATH, Loader.CUSTOMER_DATA_PATH,
             "Loader should use customer data path."],
            [lambda path: True, Loader.CUSTOMER_DATA_PATH,
             "Loader should prefer file at customer data path over built-in path."]
        ]

        for case in cases:
            loader = Loader()
            with patch('os.path.isdir') as mock_isdir:
                mock_isdir.side_effect = case[0]
                with patch.object(JSONFileLoader, 'load_file') as mock_load_file:
                    mock_load_file.side_effect = lambda path: path
                    data = loader.load_json('foo.json')
                    self.assertEquals(1, mock_load_file.call_count)
                    self.assertEquals(os.path.join(case[1], 'foo.json'), data, case[2])

    # no data dirs, load JSON
    @patch('os.path.isdir', return_value=False)
    def test_load_json_no_dirs(self, *args):
        with patch.object(JSONFileLoader, 'load_file') as mock_load_file:
            loader = Loader()
            with self.assertRaises(DataNotFoundError):
                loader.load_json('foo.json')
            self.assertEquals(0, mock_load_file.call_count)

    # no data dirs, load service model
    @patch('os.path.isdir', return_value=False)
    def test_load_service_data_no_dirs(self, *args):
        with patch.object(YAMLFileLoader, 'load_file') as mock_load_file:
            loader = Loader()
            with self.assertRaises(ValidationError):
                loader.load_service_data('service-1')
            self.assertEquals(0, mock_load_file.call_count)


# Remove (undefine) base class so that unit test runner doesn't attempt to execute.
# This skips the test without noise in output, unlike when using @unittest.skip decorator.
del TestFileLoader

if __name__ == '__main__':
    unittest.main()



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import collections
import os

from cdpcli.model import ArrayShape
from cdpcli.model import DenormalizedStructureBuilder
from cdpcli.model import InvalidShapeError
from cdpcli.model import MapShape
from cdpcli.model import NoShapeFoundError
from cdpcli.model import ObjectShape
from cdpcli.model import OperationNotFoundError
from cdpcli.model import ServiceModel
from cdpcli.model import ShapeResolver
from tests import unittest
import yaml

MODEL_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'model')


class TestServiceModel(unittest.TestCase):

    def setUp(self):
        self.model = yaml.safe_load(open(os.path.join(MODEL_DIR, 'service.yaml')))
        self.service_model = ServiceModel(self.model, 'servicename')

    def test_service_name(self):
        service_model = ServiceModel(self.model, 'myservice')
        self.assertEqual(service_model.service_name, 'myservice')

    def test_products(self):
        service_model = ServiceModel(self.model, 'myservice')
        self.assertEqual(service_model.products, ['ALTUS'])

    def test_form_factors(self):
        service_model = ServiceModel(self.model, 'myservice')
        self.assertEqual(service_model.form_factors, ['public', 'private'])

    def test_endpoint_prefix(self):
        service_model = ServiceModel(self.model, 'myservice')
        self.assertEqual(service_model.endpoint_prefix, 'api')

    def test_operations(self):
        self.assertTrue(self.service_model.operation_model(
                'createDirector') is not None)
        self.assertTrue(self.service_model.operation_model(
                'deleteDirector') is not None)
        self.assertTrue(self.service_model.operation_model(
                'describeDirectors') is not None)
        with self.assertRaises(OperationNotFoundError):
            self.service_model.operation_model('NonExistentOperation')

    def test_documentation(self):
        self.assertTrue(self.service_model.documentation.startswith(
                'Cloudera CDP Test Service Description'))

    def test_operation_names(self):
        self.assertEquals(len(self.service_model.operation_names), 3)
        self.assertTrue('createDirector' in self.service_model.operation_names)
        self.assertTrue('deleteDirector' in self.service_model.operation_names)
        self.assertTrue('describeDirectors' in self.service_model.operation_names)


class TestOperationModel(unittest.TestCase):
    def setUp(self):
        self.model = yaml.safe_load(open(os.path.join(MODEL_DIR, 'service.yaml')))
        self.service_model = ServiceModel(self.model, 'servicename')

    def test_name(self):
        operation_model = self.service_model.operation_model('createDirector')
        self.assertEqual(operation_model.name, 'createDirector')

    def test_pagination(self):
        operation_model = self.service_model.operation_model('describeDirectors')
        self.assertEqual(operation_model.can_paginate, True)
        self.assertEqual(operation_model.paging_input_token, 'startingToken')
        self.assertEqual(operation_model.paging_output_token, 'nextToken')
        self.assertEqual(operation_model.paging_result, 'directors')
        operation_model = self.service_model.operation_model('createDirector')
        self.assertEqual(operation_model.can_paginate, False)
        self.assertEqual(operation_model.paging_input_token, None)
        self.assertEqual(operation_model.paging_output_token, None)
        self.assertEqual(operation_model.paging_result, None)

    def test_form_factors(self):
        operation_model = self.service_model.operation_model('describeDirectors')
        self.assertIsNone(operation_model.form_factors)
        operation_model = self.service_model.operation_model('deleteDirector')
        self.assertEqual(operation_model.form_factors, ['public'])

    def test_operation_input(self):
        operation_model = self.service_model.operation_model('createDirector')
        self.assertEqual(operation_model.http['method'], 'post')
        self.assertEqual(operation_model.http['requestUri'], '/directors/createDirector')
        self.assertEqual(operation_model.input_shape.name, 'input')
        self.assertEqual(list(operation_model.input_shape.members), ['name'])

    def test_operation_output(self):
        operation_model = self.service_model.operation_model('createDirector')
        self.assertEqual(operation_model.output_shape.name, 'output')
        self.assertEqual(list(operation_model.output_shape.members), ['director'])

    def test_has_documentation_property(self):
        operation_model = self.service_model.operation_model('createDirector')
        self.assertEqual(operation_model.documentation, 'Creates a new Director.')

    def test_service_model_available_from_operation_model(self):
        operation_model = self.service_model.operation_model('createDirector')
        self.assertEqual(operation_model.service_model, self.service_model)


class TestShapeResolver(unittest.TestCase):
    def setUp(self):
        model = yaml.safe_load(open(os.path.join(MODEL_DIR, 'service.yaml')))
        self.resolver = ShapeResolver(model['definitions'])

    def _check_endpoint_shape(self, shape):
        self.assertEqual(shape.name, 'endpoint')
        self.assertEqual(shape.type_name, 'object')
        self.assertEqual(len(shape.required_members), 2)
        self.assertTrue('address' in shape.required_members)
        self.assertTrue('port' in shape.required_members)
        self.assertEqual(shape.members['address'].name, 'address')
        self.assertEqual(shape.members['address'].type_name, 'string')
        self.assertEqual(shape.members['port'].name, 'port')
        self.assertEqual(shape.members['port'].type_name, 'integer')

    def test_get_shape_by_name(self):
        shape = self.resolver.get_shape_by_name('endpoint', 'Endpoint')
        self._check_endpoint_shape(shape)

    def test_resolve_shape_reference(self):
        shape = self.resolver.resolve_shape_ref('endpoint', '#/definitions/Endpoint')
        self._check_endpoint_shape(shape)

    def test_min_and_max_length(self):
        shape = self.resolver.get_shape_by_name('shape', 'MinAndMaxLengthTest')
        self.assertEqual(shape.type_name, 'object')
        self.assertEqual(shape.members['stringparam'].min_length, 10)
        self.assertEqual(shape.members['stringparam'].max_length, 20)

    def test_minimum_and_maximum(self):
        shape = self.resolver.get_shape_by_name('shape', 'MinimumAndMaximumTest')
        self.assertEqual(shape.type_name, 'object')
        self.assertEqual(shape.members['intparam'].minimum, 15)
        self.assertEqual(shape.members['intparam'].maximum, 30)

    def test_list_shape(self):
        shape = self.resolver.get_shape_by_name('shape', 'DescribeDirectorsRequest')
        self.assertEqual(shape.members['names'].name, 'names')
        self.assertEqual(shape.members['names'].type_name, 'array')
        self.assertEqual(shape.members['names'].member.name, 'items')
        self.assertEqual(shape.members['names'].member.type_name, 'string')

    def test_string_shape_with_enum(self):
        shape = self.resolver.get_shape_by_name('shape', 'Director')
        self.assertEqual(shape.members['status'].name, 'status')
        self.assertEqual(shape.members['status'].type_name, 'string')
        self.assertEqual(shape.members['status'].enum, ['UNAVAILABLE',
                                                        'CREATING',
                                                        'STOPPED',
                                                        'STARTING',
                                                        'RUNNING',
                                                        'STOPPING',
                                                        'DESTROYING',
                                                        'UNKNOWN'])

    def test_no_paramfile(self):
        shape = self.resolver.get_shape_by_name('shape', 'NoParamfileTest')
        self.assertEqual(shape.type_name, 'object')
        self.assertEqual(shape.members['stringparam1'].is_no_paramfile, True)
        self.assertEqual(shape.members['stringparam2'].is_no_paramfile, False)

    def test_unknown_shape(self):
        with self.assertRaises(NoShapeFoundError):
            self.resolver.get_shape_by_name('unknown', 'Unknown')

    def test_invalid_shape_type(self):
        with self.assertRaises(InvalidShapeError):
            self.resolver.get_shape_by_name('invalid', 'MissingShapeType')
        with self.assertRaises(InvalidShapeError):
            self.resolver.get_shape_by_name('invalid', 'InvalidShapeType')

    def test_undocumented_param(self):
        shape = self.resolver.get_shape_by_name('shape', 'UndocumentedParmTest')
        self.assertEqual(shape.members['undocumentedParam'].is_undocumented, True)
        self.assertEqual(shape.members['documentedParam'].is_undocumented, False)

    def test_referenced_map(self):
        shape = self.resolver.get_shape_by_name('shape', 'ReferencedMapTest')
        self.assertEqual(shape.members['endpoints'].type_name, 'map')
        self.assertEqual(shape.members['endpoints'].key.type_name, 'string')
        self.assertEqual(shape.members['endpoints'].value.type_name, 'object')
        self.assertTrue(self._compare_lists(
            list(shape.members['endpoints'].value.members.keys()),
            ['port', 'address'])
        )

    def test_explicit_map(self):
        shape = self.resolver.get_shape_by_name('shape', 'ExplicitTypeMapTest')
        self.assertEqual(shape.members['endpoints'].type_name, 'map')
        self.assertEqual(shape.members['endpoints'].key.type_name, 'string')
        self.assertEqual(shape.members['endpoints'].value.type_name, 'object')
        self.assertTrue(self._compare_lists(
            list(shape.members['endpoints'].value.members.keys()),
            ['port', 'address'])
        )

    def test_no_freeform_map_dict_shape(self):
        with self.assertRaises(InvalidShapeError):
            shape = self.resolver.get_shape_by_name('invalid', 'NoFreeformMapDictTest')
            shape.members['endpoints'].value

    def test_no_freeform_map_true_shape(self):
        with self.assertRaises(InvalidShapeError):
            shape = self.resolver.get_shape_by_name('invalid', 'NoFreeformMapTrueTest')
            shape.members['endpoints'].value

    def test_blob_shape(self):
        shape = self.resolver.get_shape_by_name('shape', 'BlobTest')
        self.assertEqual(shape.members['blobparam'].type_name, 'blob')

    def _compare_lists(self, x, y):
        return collections.Counter(x) == collections.Counter(y)


class TestBuilders(unittest.TestCase):

    def test_structure_shape_builder_with_scalar_types(self):
        b = DenormalizedStructureBuilder()
        shape = b.with_members({
            'A': {'type': 'string'},
            'B': {'type': 'integer'},
        }).build_model()
        self.assertIsInstance(shape, ObjectShape)
        self.assertEqual(sorted(list(shape.members)), ['A', 'B'])
        self.assertEqual(shape.members['A'].type_name, 'string')
        self.assertEqual(shape.members['B'].type_name, 'integer')

    def test_structure_shape_with_object_type(self):
        b = DenormalizedStructureBuilder()
        shape = b.with_members({
            'A': {
                'type': 'object',
                'properties': {
                    'A-1': {'type': 'string'},
                }
            },
        }).build_model()
        self.assertIsInstance(shape, ObjectShape)
        self.assertEqual(list(shape.members), ['A'])
        self.assertEqual(shape.members['A'].type_name, 'object')
        self.assertEqual(list(shape.members['A'].members), ['A-1'])

    def test_structure_shape_with_list(self):
        b = DenormalizedStructureBuilder()
        shape = b.with_members({
            'A': {
                'type': 'array',
                'items': {
                    'type': 'string'
                }
            },
        }).build_model()
        self.assertIsInstance(shape.members['A'], ArrayShape)
        self.assertEqual(shape.members['A'].member.type_name, 'string')

    def test_structure_shape_with_map_type(self):
        b = DenormalizedStructureBuilder()
        shape = b.with_members({
            'A': {
                'type': 'object',
                'additionalProperties': {
                    'type': 'string'
                }
            }
        }).build_model()
        self.assertIsInstance(shape.members['A'], MapShape)
        map_shape = shape.members['A']
        self.assertEqual(map_shape.key.type_name, 'string')
        self.assertEqual(map_shape.value.type_name, 'string')



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class CLICommand(object):

    @property
    def name(self):
        raise NotImplementedError("name")

    @name.setter
    def name(self, value):
        raise NotImplementedError("name")

    @property
    def lineage(self):
        return [self]

    @property
    def lineage_names(self):
        return [cmd.name for cmd in self.lineage]

    def __call__(self, client_creator, args, parsed_globals):
        pass

    def create_help_command(self):
        return None

    @property
    def arg_table(self):
        return {}



# Copyright (c) 2021 Cloudera, Inc. All rights reserved.

from cdpcli.extensions.configure import classify
from cdpcli.extensions.configure.classify import DeploymentType
from tests import unittest


class TestClassifyDeployment(unittest.TestCase):

    def setUp(self):
        self.classifier = classify.ClassifyDeployment()

    def test_empty_endpoint_url(self):
        # Tests blank endpoint url entry during configure
        # Blank endpoint url indicates Public Cloud as per directives
        # provided to user in configure.py prompt text
        self.classifier.set_endpoint_url("")
        self.assertEqual(self.classifier.get_deployment_type(),
                         DeploymentType.PUBLIC_CLOUD)

    # Next two functions to test if URL entered is classified correctly
    # Tests function calls that return classified deployment type

    def test_classify_with_public_cloud_url(self):
        public_urls = [
            # Examples for service, env URLs
            # Public Cloud URLs follow the same format for dev/int/stage/prod(altus,cdp)
            "https://console.mow-dev.cloudera.com/",
            "http://console.thunderhead-int.cloudera.com/",
            "https://iamapi.thunderhead-stage.cloudera.com",
            "https://cloudera.cdp.mow-int.cloudera.com/",
            "https://cloudera.cdp.mow-int.cloudera.com/",
            # Example for Private Stack
            "https://console.cdp-priv.mow-dev.cloudera.com/cloud/environments/list"
            # Allow local testing via run.sh or run-with-backend.sh
            "http://localhost:8982"
        ]
        for url in public_urls:
            self.classifier.set_endpoint_url(url)
            self.assertEqual(self.classifier.get_deployment_type(),
                             DeploymentType.PUBLIC_CLOUD)
            self.assertEqual(self.classifier.is_public_cloud(), True)

    def test_classify_with_private_cloud_url(self):
        self.classifier.set_endpoint_url(
            "console-fake-priv.apps.cp-dev-02.kcloud.cloudera.com"
        )
        self.assertEqual(self.classifier.get_deployment_type(),
                         DeploymentType.PRIVATE_CLOUD)
        self.assertEqual(self.classifier.is_private_cloud(), True)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import platform
import shlex
from subprocess import PIPE
from subprocess import Popen

from cdpcli.argparser import ArgTableArgParser
from cdpcli.argprocess import ParamShorthand
from cdpcli.doc.restdoc import ReSTDocument
from cdpcli.docs import generate_doc
from cdpcli.docs import OperationDocumentGenerator
from cdpcli.docs import ProviderDocumentGenerator
from cdpcli.docs import ServiceDocumentGenerator
from cdpcli.exceptions import ExecutableNotFoundError
from cdpcli.textwriter import TextWriter
from cdpcli.utils import ignore_ctrl_c
from docutils.core import publish_string
from docutils.writers import manpage


def get_renderer():
    """
    Return the appropriate HelpRenderer implementation for the
    current platform.
    """
    if platform.system() == 'Windows':
        return WindowsHelpRenderer()
    else:
        return PosixHelpRenderer()


class NullRenderer(object):
    """
    A renderer that does not render anything. Use this when the goal is only
    to populate the help command's doc object.
    """

    def render(self, contents):
        pass


class PagingHelpRenderer(object):
    PAGER = None

    def get_pager_cmdline(self):
        pager = self.PAGER
        if 'MANPAGER' in os.environ:
            pager = os.environ['MANPAGER']
        elif 'PAGER' in os.environ:
            pager = os.environ['PAGER']
        return shlex.split(pager)

    def render(self, contents):
        """
        Each implementation of HelpRenderer must implement this
        render method.
        """
        converted_content = self._convert_doc_content(contents)
        self._send_output_to_pager(converted_content)

    def _send_output_to_pager(self, output):
        cmdline = self.get_pager_cmdline()
        p = self._popen(cmdline, stdin=PIPE)
        p.communicate(input=output)

    def _popen(self, *args, **kwargs):
        return Popen(*args, **kwargs)

    def _convert_doc_content(self, contents):
        return contents


class PosixHelpRenderer(PagingHelpRenderer):
    PAGER = 'less -R'

    def _convert_doc_content(self, contents):
        man_contents = publish_string(contents, writer=manpage.Writer())
        if not self._exists_on_path('groff'):
            raise ExecutableNotFoundError(executable_name='groff')
        cmdline = ['groff', '-man', '-T', 'ascii']
        p3 = self._popen(cmdline, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        groff_output = p3.communicate(input=man_contents)[0]
        return groff_output

    def _send_output_to_pager(self, output):
        cmdline = self.get_pager_cmdline()
        with ignore_ctrl_c():
            # We can't rely on the KeyboardInterrupt from
            # the CLIDriver being caught because when we
            # send the output to a pager it will use various
            # control characters that need to be cleaned
            # up gracefully.  Otherwise if we simply catch
            # the Ctrl-C and exit, it will likely leave the
            # users terminals in a bad state and they'll need
            # to manually run ``reset`` to fix this issue.
            # Ignoring Ctrl-C solves this issue.  It's also
            # the default behavior of less (you can't ctrl-c
            # out of a manpage).
            p = self._popen(cmdline, stdin=PIPE)
            p.communicate(input=output)

    def _exists_on_path(self, name):
        # Since we're only dealing with POSIX systems, we can
        # ignore things like PATHEXT.
        return any([os.path.exists(os.path.join(p, name))
                    for p in os.environ.get('PATH', '').split(os.pathsep)])


class WindowsHelpRenderer(PagingHelpRenderer):
    """Render help content on a Windows platform."""

    PAGER = 'more'

    def _convert_doc_content(self, contents):
        text_output = publish_string(contents,
                                     writer=TextWriter())
        return text_output

    def _popen(self, *args, **kwargs):
        # Also set the shell value to True.  To get any of the
        # piping to a pager to work, we need to use shell=True.
        kwargs['shell'] = True
        return Popen(*args, **kwargs)


class HelpCommand(object):
    GeneratorClass = None

    def __init__(self, obj, command_table, arg_table, include_man_fields=True):
        self.obj = obj
        if command_table is None:
            command_table = {}
        self.command_table = command_table
        if arg_table is None:
            arg_table = {}
        self.arg_table = arg_table
        self.include_man_fields = include_man_fields
        self._subcommand_table = {}
        self._related_items = []
        self.renderer = get_renderer()
        self.doc = ReSTDocument(target='man')

    @property
    def command_lineage(self):
        pass

    @property
    def name(self):
        pass

    @property
    def subcommand_table(self):
        return self._subcommand_table

    @property
    def related_items(self):
        return self._related_items

    def __call__(self, client_creator, args, parsed_globals):
        if args:
            subcommand_parser = ArgTableArgParser({}, command_table=self.subcommand_table)
            parsed, remaining = subcommand_parser.parse_known_args(args)
            if getattr(parsed, 'subcommand', None) is not None:
                return self.subcommand_table[parsed.subcommand](remaining,
                                                                parsed_globals)

        generate_doc(self.GeneratorClass(self), self)
        self.renderer.render(self.doc.getvalue())


class ProviderHelpCommand(HelpCommand):
    GeneratorClass = ProviderDocumentGenerator

    def __init__(self, command_table, arg_table, description, synopsis, usage):
        HelpCommand.__init__(self, None, command_table, arg_table)
        self.description = description
        self.synopsis = synopsis
        self.help_usage = usage

    @property
    def command_lineage(self):
        return 'cdp'

    @property
    def name(self):
        return 'cdp'


class ServiceHelpCommand(HelpCommand):
    GeneratorClass = ServiceDocumentGenerator

    def __init__(self, obj, command_table, arg_table, name, command_lineage):
        super(ServiceHelpCommand, self).__init__(obj, command_table, arg_table)
        self._name = name
        self._command_lineage = command_lineage

    @property
    def command_lineage(self):
        return self._command_lineage

    @property
    def name(self):
        return self._name


class OperationHelpCommand(HelpCommand):
    GeneratorClass = OperationDocumentGenerator

    def __init__(self, operation_model, arg_table, name, command_lineage):
        HelpCommand.__init__(self, operation_model, None, arg_table)
        self.param_shorthand = ParamShorthand()
        self._name = name
        self._command_lineage = command_lineage

    @property
    def command_lineage(self):
        return self._command_lineage

    @property
    def name(self):
        return self._name



# Copyright (c) 2016 Cloudera, Inc. All rights reserved.

import os

from cdpcli.argparser import ArgTableArgParser
from cdpcli.argparser import ServiceArgParser
from cdpcli.clidriver import CLIDriver, ServiceOperation
from cdpcli.model import ServiceModel
from mock import Mock
from tests import unittest
import yaml

MODEL_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                         'argparser')
TEST_ARGS = ['submit-spark-job',
             '--jars',
             's3a://foobar/barfoo-1.0.jar',
             u'\\-s3a://foobar/barfoo-1.0.jar',
             '--main-class',
             'com.foobarcorp.big_data_maker.SimpleApp',
             '--cluster-name=Cluster9',
             '--arguments',
             '\\-o',
             's3a://foobar-demo/big.out/',
             u'\\-p',
             '100',
             '\\-rfoo\\-bar\\-',
             '1000000']


class TestServiceArgParser(unittest.TestCase):

    def setUp(self):
        self.parser = ServiceArgParser({'submit-spark-job': ''}, 'dataeng')
        self.args = TEST_ARGS

    def testOperationDetected(self):
        parsed, remaining = self.parser.parse_known_args(self.args)
        self.assertEqual('submit-spark-job', parsed.operation)
        self.assertEqual(self.args[1:], remaining)


class TestArgTableArgParser(unittest.TestCase):

    def setUp(self):
        self.args = TEST_ARGS[1:]
        model = yaml.safe_load(open(os.path.join(MODEL_DIR, 'service.yaml')))
        service_model = ServiceModel(model, 'servicename')
        service_operation = ServiceOperation(
            Mock(spec=CLIDriver),
            'submit-spark-job',
            'dataeng',
            '',
            service_model.operation_model('submitSparkJob'))
        self.parser = ArgTableArgParser(service_operation.arg_table)

    def testArsParsedCorrectly(self):
        parsed, remaining = self.parser.parse_known_args(self.args)
        self.assertEqual([], remaining)
        parsed_args = vars(parsed)
        # The second jar starts with a dash. We should decode it correctly.
        self.assertEquals(['s3a://foobar/barfoo-1.0.jar',
                           '-s3a://foobar/barfoo-1.0.jar'],
                          parsed_args['jars'])
        self.assertEquals('com.foobarcorp.big_data_maker.SimpleApp',
                          parsed_args['main_class'])
        self.assertEquals('Cluster9', parsed_args['cluster_name'])
        # The following tests that we decoded the starting dash
        self.assertEquals(
            ['-o',
             's3a://foobar-demo/big.out/',
             '-p',
             '100',
             '-rfoo\\-bar\\-',
             '1000000'],
            parsed_args['arguments'])
